import base64
import collections
import itertools
import os
import math
import struct
import pprint # pylint: disable=unused-import

from panda3d.core import * # pylint: disable=wildcard-import
try:
    from panda3d import bullet
    HAVE_BULLET = True
except ImportError:
    HAVE_BULLET = False

load_prc_file_data(
    __file__,
    'interpolate-frames #t\n'
)


class Converter():
    _COMPONENT_TYPE_MAP = {
        5120: GeomEnums.NT_int8,
        5121: GeomEnums.NT_uint8,
        5122: GeomEnums.NT_int16,
        5123: GeomEnums.NT_uint16,
        5124: GeomEnums.NT_int32,
        5125: GeomEnums.NT_uint32,
        5126: GeomEnums.NT_float32,
    }
    _COMPONENT_NUM_MAP = {
        'MAT4': 16,
        'VEC4': 4,
        'VEC3': 3,
        'VEC2': 2,
        'SCALAR': 1,
    }
    _ATTRIB_CONENT_MAP = {
        'vertex': GeomEnums.C_point,
        'normal': GeomEnums.C_normal,
        'texcoord': GeomEnums.C_texcoord,
        'color': GeomEnums.C_color,
        'weights': GeomEnums.C_point,
        'joints': GeomEnums.C_point,
    }
    _ATTRIB_NAME_MAP = {
        'position': InternalName.get_vertex().get_name(),
        'weights': InternalName.get_transform_weight().get_name(),
        'joints': InternalName.get_transform_index().get_name(),
    }
    _PRIMITIVE_MODE_MAP = {
        0: GeomPoints,
        1: GeomLines,
        3: GeomLinestrips,
        4: GeomTriangles,
        5: GeomTristrips,
        6: GeomTrifans,
    }

    def __init__(self, indir=Filename.from_os_specific(os.getcwd()), outdir=Filename.from_os_specific(os.getcwd())):
        self.indir = indir
        self.outdir = outdir
        self.cameras = {}
        self.buffers = {}
        self.lights = {}
        self.textures = {}
        self.mat_states = {}
        self.mat_mesh_map = {}
        self.meshes = {}
        self.nodes = {}
        self.node_paths = {}
        self.scenes = {}
        self.characters = {}
        self.joint_map = {}

        # Coordinate system transform matrix
        self.csxform = LMatrix4.convert_mat(CS_yup_right, CS_default)
        self.csxform_inv = LMatrix4.convert_mat(CS_default, CS_yup_right)

        self._joint_nodes = set()

        # Scene props
        self.active_scene = NodePath(ModelRoot('default'))
        self.background_color = (0, 0, 0)
        self.active_camera = None

    def update(self, gltf_data, writing_bam=False):
        #pprint.pprint(gltf_data)

        if 'extensionsUsed' in gltf_data and 'BP_zup' in gltf_data['extensionsUsed']:
            self.csxform = LMatrix4.ident_mat()
            self.csxform_inv = LMatrix4.ident_mat()

        # Convert data
        for buffid, gltf_buffer in enumerate(gltf_data.get('buffers', [])):
            self.load_buffer(buffid, gltf_buffer)

        for camid, gltf_cam in enumerate(gltf_data.get('cameras', [])):
            self.load_camera(camid, gltf_cam)

        if 'extensions' in gltf_data and 'KHR_lights' in gltf_data['extensions']:
            lights = gltf_data['extensions']['KHR_lights'].get('lights', [])
            for lightid, gltf_light in enumerate(lights):
                self.load_light(lightid, gltf_light)

        for texid, gltf_tex in enumerate(gltf_data.get('textures', [])):
            self.load_texture(texid, gltf_tex, gltf_data)
        self.load_fallback_texture()

        for matid, gltf_mat in enumerate(gltf_data.get('materials', [])):
            self.load_material(matid, gltf_mat)

        for skinid, gltf_skin in enumerate(gltf_data.get('skins', [])):
            self.load_skin(skinid, gltf_skin, gltf_data)

        for meshid, gltf_mesh in enumerate(gltf_data.get('meshes', [])):
            self.load_mesh(meshid, gltf_mesh, gltf_data)

        for nodeid, gltf_node in enumerate(gltf_data.get('nodes', [])):
            node_name = gltf_node.get('name', 'node'+str(nodeid))
            node = self.nodes.get(nodeid, PandaNode(node_name))
            self.nodes[nodeid] = node

        # If we support writing bam 6.40, we can safely write out
        # instanced lights.  If not, we have to copy it.
        copy_lights = writing_bam and not hasattr(BamWriter, 'root_node')

        # Build scenegraphs
        def add_node(root, gltf_scene, nodeid):
            try:
                gltf_node = gltf_data['nodes'][nodeid]
            except IndexError:
                print("Could not find node with index: {}".format(nodeid))
                return

            node_name = gltf_node.get('name', 'node'+str(nodeid))
            if nodeid in self._joint_nodes:
                # don't handle joints here
                return
            panda_node = self.nodes[nodeid]

            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
                if nodeid in gltf_scene['extras']['hidden_nodes']:
                    panda_node = panda_node.make_copy()

            np = self.node_paths.get(nodeid, root.attach_new_node(panda_node))
            self.node_paths[nodeid] = np

            if 'mesh' in gltf_node:
                mesh = self.meshes[gltf_node['mesh']]
                np_tmp = np

                if 'skin' in gltf_node:
                    char = self.characters[gltf_node['skin']]
                    np_tmp = np.attach_new_node(char)

                    self.combine_mesh_skin(mesh, gltf_node['skin'])
                np_tmp.attach_new_node(mesh)
            if 'skin' in gltf_node and not 'mesh' in gltf_node:
                print(
                    "Warning: node {} has a skin but no mesh"
                    .format(primitiveid)
                )
            if 'camera' in gltf_node:
                camid = gltf_node['camera']
                cam = self.cameras[camid]
                np.attach_new_node(cam)
            if 'extensions' in gltf_node:
                if 'KHR_lights' in gltf_node['extensions']:
                    lightid = gltf_node['extensions']['KHR_lights']['light']
                    light = self.lights[lightid]
                    if copy_lights:
                        light = light.make_copy()
                    lnp = np.attach_new_node(light)
                    if isinstance(light, Light):
                        root.set_light(lnp)

                if HAVE_BULLET and 'BLENDER_physics' in gltf_node['extensions']:
                    phy = gltf_node['extensions']['BLENDER_physics']
                    shape = None
                    collision_shape = phy['collisionShapes'][0]
                    bounding_box = collision_shape['boundingBox']
                    radius = max(bounding_box[0], bounding_box[1]) / 2.0
                    height = bounding_box[2]
                    geomnode = None
                    static = 'static' in phy and phy['static']
                    if 'mesh' in collision_shape:
                        try:
                            geomnode = self.meshes[collision_shape['mesh']]
                        except KeyError:
                            print(
                                "Could not find physics mesh ({}) for object ({})"
                                .format(collision_shape['mesh'], nodeid)
                            )

                    shape_type = collision_shape['shapeType']
                    if shape_type == 'BOX':
                        shape = bullet.BulletBoxShape(LVector3(*bounding_box) / 2.0)
                    elif shape_type == 'SPHERE':
                        shape = bullet.BulletSphereShape(max(bounding_box) / 2.0)
                    elif shape_type == 'CAPSULE':
                        shape = bullet.BulletCapsuleShape(radius, height - 2.0 * radius, bullet.ZUp)
                    elif shape_type == 'CYLINDER':
                        shape = bullet.BulletCylinderShape(radius, height, bullet.ZUp)
                    elif shape_type == 'CONE':
                        shape = bullet.BulletConeShape(radius, height, bullet.ZUp)
                    elif shape_type == 'CONVEX_HULL':
                        if geomnode:
                            shape = bullet.BulletConvexHullShape()

                            for geom in geomnode.get_geoms():
                                shape.add_geom(geom)
                    elif shape_type == 'MESH':
                        if geomnode:
                            mesh = bullet.BulletTriangleMesh()
                            for geom in geomnode.get_geoms():
                                mesh.add_geom(geom)
                            shape = bullet.BulletTriangleMeshShape(mesh, dynamic=not static)
                    else:
                        print("Unknown collision shape ({}) for object ({})".format(shape_type, nodeid))

                    if shape is not None:
                        phynode = bullet.BulletRigidBodyNode(node_name)
                        phynode.add_shape(shape)
                        np.attach_new_node(phynode)
                        if not static:
                            phynode.set_mass(phy['mass'])
                    else:
                        print("Could not create collision shape for object ({})".format(nodeid))
                elif not HAVE_BULLET:
                    print("Bullet is unavailable, not converting collision shape for object ({})".format(nodeid))
            if 'extras' in gltf_node:
                for key, value in gltf_node['extras'].items():
                    np.set_tag(key, str(value))


            for child_nodeid in gltf_node.get('children', []):
                add_node(np, gltf_scene, child_nodeid)

            # Handle visibility after children are loaded
            def visible_recursive(node, visible):
                if visible:
                    node.show()
                else:
                    node.hide()
                for child in node.get_children():
                    visible_recursive(child, visible)
            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
                if nodeid in gltf_scene['extras']['hidden_nodes']:
                    #print('Hiding', np)
                    visible_recursive(np, False)
                else:
                    #print('Showing', np)
                    visible_recursive(np, True)

            # Check if we need to deal with negative scale values
            scale = panda_node.get_transform().get_scale()
            negscale = scale.x * scale.y * scale.z < 0
            if negscale:
                for geomnode in np.find_all_matches('**/+GeomNode'):
                    tmp = geomnode.get_parent().attach_new_node(PandaNode('ReverseCulling'))
                    tmp.set_attrib(CullFaceAttrib.make_reverse())
                    geomnode.reparent_to(tmp)

        for sceneid, gltf_scene in enumerate(gltf_data.get('scenes', [])):
            scene_name = gltf_scene.get('name', 'scene'+str(sceneid))
            scene_root = NodePath(ModelRoot(scene_name))

            node_list = gltf_scene['nodes']
            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
                node_list += gltf_scene['extras']['hidden_nodes']

            for nodeid in node_list:
                add_node(scene_root, gltf_scene, nodeid)

            self.scenes[sceneid] = scene_root

        # Update node transforms for glTF nodes that have a NodePath
        for nodeid, gltf_node in enumerate(gltf_data.get('nodes', [])):
            if nodeid not in self.node_paths:
                continue
            np = self.node_paths[nodeid]

            gltf_pos = LVector3(*gltf_node.get('translation', [0, 0, 0]))
            gltf_rot = self.load_quaternion_as_hpr(gltf_node.get('rotation', [0, 0, 0, 1]))
            gltf_scale = LVector3(*gltf_node.get('scale', [1, 1, 1]))

            gltf_mat = LMatrix4()
            compose_matrix(gltf_mat, gltf_scale, gltf_rot, gltf_pos, CS_yup_right)
            if np.has_parent():
                parent_mat = np.get_parent().get_mat()
            else:
                parent_mat = LMatrix4.ident_mat()
            parent_inv = LMatrix4(parent_mat)
            parent_inv.invert_in_place()
            np.set_mat(gltf_mat * parent_inv * self.csxform)
            # Switch to this line when mesh axis-transform works
            #np.set_mat(self.csxform * gltf_mat * self.csxform_inv)

        # Set the active scene
        sceneid = gltf_data.get('scene', 0)
        if sceneid in self.scenes:
            self.active_scene = self.scenes[sceneid]
        if 'scenes' in gltf_data:
            gltf_scene = gltf_data['scenes'][sceneid]
            if 'extras' in gltf_scene:
                if 'background_color' in gltf_scene['extras']:
                    self.background_color = gltf_scene['extras']['background_color']
                if 'active_camera' in gltf_scene['extras']:
                    self.active_camera = gltf_scene['extras']['active_camera']

    def load_matrix(self, mat):
        lmat = LMatrix4()

        for i in range(4):
            lmat.set_row(i, LVecBase4(*mat[i * 4: i * 4 + 4]))
        return lmat

    def decompose_matrix(self, mat):
        mat = LMatrix4(mat)
        translation = mat.get_row3(3)
        mat.set_row(3, LVector3(0, 0, 0))
        scale = [mat.get_row3(i).length() for i in range(3)]
        for i in range(3):
            mat.set_row(i, mat.get_row3(i) / scale[i])
        rot_quat = LQuaternion()
        rot_quat.set_from_matrix(mat.getUpper3())
        rotation = rot_quat.get_hpr()

        return translation, rotation, LVector3(*scale)

    def load_quaternion_as_hpr(self, quaternion):
        quat = LQuaternion(quaternion[3], quaternion[0], quaternion[1], quaternion[2])
        return quat.get_hpr()

    def load_buffer(self, buffid, gltf_buffer):
        uri = gltf_buffer['uri']
        if uri.startswith('data:application/octet-stream;base64'):
            buff_data = gltf_buffer['uri'].split(',')[1]
            buff_data = base64.b64decode(buff_data)
        elif uri.endswith('.bin'):
            buff_fname = os.path.join(self.indir.to_os_specific(), uri)
            with open(buff_fname, 'rb') as buff_file:
                buff_data = buff_file.read()
        else:
            print(
                "Buffer {} has an unsupported uri ({}), using a zero filled buffer instead"
                .format(buffid, uri)
            )
            buff_data = bytearray(gltf_buffer['byteLength'])
        self.buffers[buffid] = buff_data

    def make_texture_srgb(self, texture):
        if texture.get_num_components() == 3:
            texture.set_format(Texture.F_srgb)
        elif texture.get_num_components() == 4:
            texture.set_format(Texture.F_srgb_alpha)

    def load_fallback_texture(self):
        texture = Texture('pbr-fallback')
        texture.setup_2d_texture(1, 1, Texture.T_unsigned_byte, Texture.F_rgba)
        texture.set_clear_color(LColor(1, 1, 1, 1))

        self.textures['__bp-pbr-fallback'] = texture

    def load_texture(self, texid, gltf_tex, gltf_data):
        if 'source' not in gltf_tex:
            print("Texture '{}' has no source, skipping".format(texid))
            return

        source = gltf_data['images'][gltf_tex['source']]
        uri = source['uri']
        def write_tex_image(ext):
            texname = 'tex{}.{}'.format(gltf_tex['source'], ext)
            texdata = base64.b64decode(uri.split(',')[1])
            texfname = os.path.join(self.outdir.to_os_specific(), texname)
            with open(texfname, 'wb') as texfile:
                texfile.write(texdata)
            return texfname
        if uri.startswith('data:image/png;base64'):
            uri = write_tex_image('png')
        elif uri.startswith('data:image/jpeg;base64'):
            uri = write_tex_image('jpeg')
        else:
            uri = Filename.fromOsSpecific(uri)
        texture = TexturePool.load_texture(uri, 0, False, LoaderOptions())

        if 'sampler' in gltf_tex:
            gltf_sampler = gltf_data['samplers'][gltf_tex['sampler']]
            if 'magFilter' in gltf_sampler:
                if gltf_sampler['magFilter'] == 9728:
                    texture.set_magfilter(SamplerState.FT_nearest)
                elif gltf_sampler['magFilter'] == 9729:
                    texture.set_magfilter(SamplerState.FT_linear)
                else:
                    print(
                        "Sampler {} has unsupported magFilter type {}"
                        .format(gltf_tex['sampler'], gltf_sampler['magFilter'])
                    )
            if 'minFilter' in gltf_sampler:
                if gltf_sampler['minFilter'] == 9728:
                    texture.set_minfilter(SamplerState.FT_nearest)
                elif gltf_sampler['minFilter'] == 9729:
                    texture.set_minfilter(SamplerState.FT_linear)
                elif gltf_sampler['minFilter'] == 9984:
                    texture.set_minfilter(SamplerState.FT_nearest_mipmap_nearest)
                elif gltf_sampler['minFilter'] == 9985:
                    texture.set_minfilter(SamplerState.FT_linear_mipmap_nearest)
                elif gltf_sampler['minFilter'] == 9986:
                    texture.set_minfilter(SamplerState.FT_nearest_mipmap_linear)
                elif gltf_sampler['minFilter'] == 9987:
                    texture.set_minfilter(SamplerState.FT_linear_mipmap_linear)
                else:
                    print(
                        "Sampler {} has unsupported minFilter type {}"
                        .format(gltf_tex['sampler'], gltf_sampler['minFilter'])
                    )

            wraps = gltf_sampler.get('wrapS', 10497)
            if wraps == 33071:
                texture.set_wrap_u(SamplerState.WM_clamp)
            elif wraps == 33648:
                texture.set_wrap_u(SamplerState.WM_mirror)
            elif wraps == 10497:
                texture.set_wrap_u(SamplerState.WM_repeat)
            else:
                print(
                    "Sampler {} has unsupported wrapS type {}"
                    .format(gltf_tex['sampler'], gltf_sampler['wrapS'])
                )

            wrapt = gltf_sampler.get('wrapT', 10497)
            if wrapt == 33071:
                texture.set_wrap_v(SamplerState.WM_clamp)
            elif wrapt == 33648:
                texture.set_wrap_v(SamplerState.WM_mirror)
            elif wrapt == 10497:
                texture.set_wrap_v(SamplerState.WM_repeat)
            else:
                print(
                    "Sampler {} has unsupported wrapT type {}"
                    .format(gltf_tex['sampler'], gltf_sampler['wrapT'])
                )

        self.textures[texid] = texture

    def load_material(self, matid, gltf_mat):
        matname = gltf_mat.get('name', 'mat'+str(matid))
        state = self.mat_states.get(matid, RenderState.make_empty())

        if matid not in self.mat_mesh_map:
            self.mat_mesh_map[matid] = []

        pmat = Material(matname)
        pbr_fallback = {'index': '__bp-pbr-fallback', 'texCoord': 0}
        texinfos = []

        if 'extensions' in gltf_mat and 'BP_materials_legacy' in gltf_mat['extensions']:
            matsettings = gltf_mat['extensions']['BP_materials_legacy']['bpLegacy']
            pmat.set_shininess(matsettings['shininessFactor'])
            pmat.set_ambient(LColor(*matsettings['ambientFactor']))

            if 'diffuseTexture' in matsettings:
                texinfo = matsettings['diffuseTexture']
                texinfos.append(texinfo)
                if matsettings['diffuseTextureSrgb'] and texinfo['index'] in self.textures:
                    self.make_texture_srgb(self.textures[texinfo['index']])
            else:
                pmat.set_diffuse(LColor(*matsettings['diffuseFactor']))

            if 'emissionTexture' in matsettings:
                texinfo = matsettings['emissionTexture']
                texinfos.append(texinfo)
                if matsettings['emissionTextureSrgb'] and texinfo['index'] in self.textures:
                    self.make_texture_srgb(self.textures[texinfo['index']])
            else:
                pmat.set_emission(LColor(*matsettings['emissionFactor']))

            if 'specularTexture' in matsettings:
                texinfo = matsettings['specularTexture']
                texinfos.append(texinfo)
                if matsettings['specularTextureSrgb'] and texinfo['index'] in self.textures:
                    self.make_texture_srgb(self.textures[texinfo['index']])
            else:
                pmat.set_specular(LColor(*matsettings['specularFactor']))
        elif 'pbrMetallicRoughness' in gltf_mat:
            pbrsettings = gltf_mat['pbrMetallicRoughness']

            pmat.set_base_color(LColor(*pbrsettings.get('baseColorFactor', [1.0, 1.0, 1.0, 1.0])))
            texinfos.append(pbrsettings.get('baseColorTexture', pbr_fallback))
            if texinfos[-1]['index'] in self.textures:
                self.make_texture_srgb(self.textures[texinfos[-1]['index']])

            pmat.set_metallic(pbrsettings.get('metallicFactor', 1.0))
            pmat.set_roughness(pbrsettings.get('roughnessFactor', 1.0))
            texinfos.append(pbrsettings.get('metallicRoughnessTexture', pbr_fallback))

        pmat.set_twoside(gltf_mat.get('doubleSided', False))

        state = state.set_attrib(MaterialAttrib.make(pmat))

        # Setup textures
        tex_attrib = TextureAttrib.make()
        for i, texinfo in enumerate(texinfos):
            texdata = self.textures.get(texinfo['index'], None)
            if texdata is None:
                print("Could not find texture for key: {}".format(texinfo['index']))
                continue

            texstage = TextureStage(str(i))
            texstage.set_texcoord_name(InternalName.get_texcoord_name(str(texinfo.get('texCoord', 0))))
            tex_attrib = tex_attrib.add_on_stage(texstage, texdata)

        state = state.set_attrib(tex_attrib)

        # Setup Alpha mode
        alpha_mode = gltf_mat.get('alphaMode', 'OPAQUE')
        if alpha_mode == 'MASK':
            alpha_cutoff = gltf_mat.get('alphaCutoff', 0.5)
            alpha_attrib = AlphaTestAttrib.make(AlphaTestAttrib.M_greater_equal, alpha_cutoff)
            state = state.set_attrib(alpha_attrib)
        elif alpha_mode == 'BLEND':
            transp_attrib = TransparencyAttrib.make(TransparencyAttrib.M_alpha)
            state = state.set_attrib(transp_attrib)
        elif alpha_mode != 'OPAQUE':
            print(
                "Warning: material {} has an unsupported alphaMode: {}"
                .format(matid, alpha_mode)
            )

        # Remove stale meshes
        self.mat_mesh_map[matid] = [
            pair for pair in self.mat_mesh_map[matid] if pair[0] in self.meshes
        ]

        # Reload the material
        for meshid, geom_idx in self.mat_mesh_map[matid]:
            self.meshes[meshid].set_geom_state(geom_idx, state)

        self.mat_states[matid] = state

    def create_anim(self, character, root_bone_id, animid, gltf_anim, gltf_data):
        anim_name = gltf_anim.get('name', 'anim'+str(animid))
        samplers = gltf_anim['samplers']

        # Blender exports the same number of elements in each time parameter, so find
        # one and assume that the number of elements is the number of frames
        time_acc_id = samplers[0]['input']
        time_acc = gltf_data['accessors'][time_acc_id]
        time_bv = gltf_data['bufferViews'][time_acc['bufferView']]
        start = time_acc['byteOffset'] + time_bv['byteOffset']
        end = start + time_acc['count'] * 4
        time_data = [
            struct.unpack_from('<f', self.buffers[time_bv['buffer']], idx)[0]
            for idx in range(start, end, 4)
        ]
        fps = 1 / time_data[-1]
        num_frames = time_acc['count']

        bundle_name = anim_name
        bundle = AnimBundle(bundle_name, fps, num_frames)
        skeleton = AnimGroup(bundle, '<skeleton>')

        def create_anim_channel(parent, boneid):
            bone = gltf_data['nodes'][boneid]
            bone_name = bone.get('name', 'bone'+str(boneid))
            channels = [chan for chan in gltf_anim['channels'] if chan['target']['node'] == boneid]
            joint_mat = character.find_joint(bone_name).get_transform()

            group = AnimChannelMatrixXfmTable(parent, bone_name)

            def get_accessor(path):
                accessors = [
                    gltf_data['accessors'][samplers[chan['sampler']]['output']]
                    for chan in channels
                    if chan['target']['path'] == path
                ]

                return accessors[0] if accessors else None

            def extract_chan_data(path):
                vals = []
                acc = get_accessor(path)

                buff_view = gltf_data['bufferViews'][acc['bufferView']]
                buff_data = self.buffers[buff_view['buffer']]
                start = acc['byteOffset'] + buff_view['byteOffset']

                if path == 'rotation':
                    end = start + acc['count'] * 4 * 4
                    data = [struct.unpack_from('<ffff', buff_data, idx) for idx in range(start, end, 4 * 4)]
                    vals = [
                        [i[0] for i in data],
                        [i[1] for i in data],
                        [i[2] for i in data],
                        [i[3] for i in data]
                    ]
                    #convert quats to hpr
                    vals = list(zip(*[LQuaternion(i[3], i[0], i[1], i[2]).get_hpr() for i in zip(*vals)]))
                else:
                    end = start + acc['count'] * 3 * 4
                    data = [struct.unpack_from('<fff', buff_data, idx) for idx in range(start, end, 3 * 4)]
                    vals = [
                        [i[0] for i in data],
                        [i[1] for i in data],
                        [i[2] for i in data]
                    ]

                return vals

            # Create default animaton data
            translation, rotation, scale = self.decompose_matrix(joint_mat)
            loc_vals = list(zip(
                *[(translation.get_x(), translation.get_y(), translation.get_z()) for i in range(num_frames)]
            ))
            rot_vals = list(zip(
                *[(rotation.get_x(), rotation.get_y(), rotation.get_z()) for i in range(num_frames)]
            ))
            scale_vals = list(zip(
                *[(scale.get_x(), scale.get_y(), scale.get_z()) for i in range(num_frames)]
            ))

            # Override defaults with any found animation data
            if get_accessor('translation') is not None:
                loc_vals = extract_chan_data('translation')
            if get_accessor('rotation') is not None:
                rot_vals = extract_chan_data('rotation')
            if get_accessor('scale') is not None:
                scale_vals = extract_chan_data('scale')

            # Write data to tables
            group.set_table(b'x', CPTAFloat(PTAFloat(loc_vals[0])))
            group.set_table(b'y', CPTAFloat(PTAFloat(loc_vals[1])))
            group.set_table(b'z', CPTAFloat(PTAFloat(loc_vals[2])))

            group.set_table(b'h', CPTAFloat(PTAFloat(rot_vals[0])))
            group.set_table(b'p', CPTAFloat(PTAFloat(rot_vals[1])))
            group.set_table(b'r', CPTAFloat(PTAFloat(rot_vals[2])))

            group.set_table(b'i', CPTAFloat(PTAFloat(scale_vals[0])))
            group.set_table(b'j', CPTAFloat(PTAFloat(scale_vals[1])))
            group.set_table(b'k', CPTAFloat(PTAFloat(scale_vals[2])))

            for childid in bone.get('children', []):
                create_anim_channel(group, childid)

        create_anim_channel(skeleton, root_bone_id)
        character.add_child(AnimBundleNode(character.name, bundle))

    def load_skin(self, skinid, gltf_skin, gltf_data):
        skinname = gltf_skin.get('name', 'char'+str(skinid))
        #print("Creating character for", skinname)
        root = gltf_data['nodes'][gltf_skin['skeleton']]

        character = Character(skinname)
        bundle = character.get_bundle(0)
        skeleton = PartGroup(bundle, "<skeleton>")
        jvtmap = {}

        bind_mats = []
        ibmacc = gltf_data['accessors'][gltf_skin['inverseBindMatrices']]
        ibmbv = gltf_data['bufferViews'][ibmacc['bufferView']]
        start = ibmacc['byteOffset'] + ibmbv['byteOffset']
        end = start + ibmacc['count'] * 16 * 4
        ibmdata = self.buffers[ibmbv['buffer']][start:end]

        joint_ids = set()

        for i in range(ibmacc['count']):
            mat = struct.unpack_from('<{}'.format('f'*16), ibmdata, i * 16 * 4)
            #print('loaded', mat)
            mat = self.load_matrix(mat)
            mat.invert_in_place()
            bind_mats.append(mat)

        def create_joint(parent, nodeid, node, transform):
            node_name = node.get('name', 'bone'+str(nodeid))
            inv_transform = LMatrix4(transform)
            inv_transform.invert_in_place()
            joint_index = None
            joint_mat = LMatrix4.ident_mat()
            if nodeid in gltf_skin['joints']:
                joint_index = gltf_skin['joints'].index(nodeid)
                joint_mat = bind_mats[joint_index]
                self._joint_nodes.add(nodeid)

            # glTF uses an absolute bind pose, Panda wants it local
            bind_pose = joint_mat * inv_transform
            joint = CharacterJoint(character, bundle, parent, node_name, bind_pose)

            # Non-deforming bones are not in the skin's jointNames, don't add them to the jvtmap
            if joint_index is not None:
                jvtmap[joint_index] = JointVertexTransform(joint)

            joint_ids.add(nodeid)

            for child in node.get('children', []):
                #print("Create joint for child", child)
                bone_node = gltf_data['nodes'][child]
                create_joint(joint, child, bone_node, bind_pose * transform)

        create_joint(skeleton, gltf_skin['skeleton'], root, LMatrix4.ident_mat())

        self.characters[skinid] = character
        self.joint_map[skinid] = jvtmap

        # convert animations
        #print("Looking for actions for", skinname, joint_ids)
        anims = [
            (animid, anim)
            for animid, anim in enumerate(gltf_data.get('animations', []))
            if joint_ids & {chan['target']['node'] for chan in anim['channels']}
        ]

        if anims:
            #print("Found anims for", skinname)
            for animid, gltf_anim in anims:
                #print("\t", gltf_anim.get('name', 'anim'+str(animid)))
                self.create_anim(character, gltf_skin['skeleton'], animid, gltf_anim, gltf_data)

    def load_primitive(self, geom_node, gltf_primitive, gltf_data):
        # Build Vertex Format
        vformat = GeomVertexFormat()
        mesh_attribs = gltf_primitive['attributes']
        accessors = [
            {**gltf_data['accessors'][acc_idx], 'attrib': attrib_name}
            for attrib_name, acc_idx in mesh_attribs.items()
        ]
        accessors = sorted(accessors, key=lambda x: x['bufferView'])
        data_copies = []
        is_skinned = 'JOINTS_0' in mesh_attribs

        for buffview, accs in itertools.groupby(accessors, key=lambda x: x['bufferView']):
            buffview = gltf_data['bufferViews'][buffview]
            accs = sorted(accs, key=lambda x: x.get('byteOffset', 0))
            is_interleaved = len(accs) > 1 and accs[1]['byteOffset'] < buffview['byteStride']

            varray = GeomVertexArrayFormat()
            for acc in accs:
                # Gather column information
                attrib_parts = acc['attrib'].lower().split('_')
                attrib_name = self._ATTRIB_NAME_MAP.get(attrib_parts[0], attrib_parts[0])
                if attrib_name == 'texcoord' and len(attrib_parts) > 1:
                    internal_name = InternalName.make(attrib_name+'.', int(attrib_parts[1]))
                else:
                    internal_name = InternalName.make(attrib_name)
                num_components = self._COMPONENT_NUM_MAP[acc['type']]
                numeric_type = self._COMPONENT_TYPE_MAP[acc['componentType']]
                content = self._ATTRIB_CONENT_MAP.get(attrib_name, GeomEnums.C_other)

                # Add this accessor as a column to the current vertex array format
                varray.add_column(internal_name, num_components, numeric_type, content)

                if not is_interleaved:
                    # Start a new vertex array format
                    vformat.add_array(varray)
                    varray = GeomVertexArrayFormat()
                    data_copies.append((
                        buffview['buffer'],
                        acc.get('byteOffset', 0) + buffview.get('byteOffset', 0),
                        acc['count'],
                        buffview.get('byteStride', 4 * num_components)
                    ))

            if is_interleaved:
                vformat.add_array(varray)
                data_copies.append((
                    buffview['buffer'],
                    buffview['byteOffset'],
                    accs[0]['count'],
                    buffview.get('byteStride', varray.get_stride())
                ))

        # Copy data from buffers
        reg_format = GeomVertexFormat.register_format(vformat)
        vdata = GeomVertexData(geom_node.name, reg_format, GeomEnums.UH_stream)

        for array_idx, data_info in enumerate(data_copies):
            handle = vdata.modify_array(array_idx).modify_handle()
            handle.unclean_set_num_rows(data_info[2])

            buff = self.buffers[data_info[0]]
            start = data_info[1]
            end = start + data_info[2] * data_info[3]
            handle.copy_data_from(buff[start:end])
            handle = None

        # Flip UVs
        num_uvs = len({i for i in gltf_primitive['attributes'] if i.startswith('TEXCOORD')})
        for i in range(num_uvs):
            uv_data = GeomVertexRewriter(vdata, InternalName.get_texcoord_name(str(i)))

            while not uv_data.is_at_end():
                uvs = uv_data.get_data2f()
                uv_data.set_data2f(uvs[0], 1 - uvs[1])

        # Repack mesh data
        vformat = GeomVertexFormat()
        varray_vert = GeomVertexArrayFormat()
        varray_skin = GeomVertexArrayFormat()

        skip_columns = (
            InternalName.get_transform_index(),
            InternalName.get_transform_weight(),
            InternalName.get_transform_blend()
        )
        for arr in reg_format.get_arrays():
            for column in arr.get_columns():
                varray = varray_skin if column.get_name() in skip_columns else varray_vert
                varray.add_column(
                    column.get_name(),
                    column.get_num_components(),
                    column.get_numeric_type(),
                    column.get_contents()
                )
        vformat.add_array(varray_vert)

        if is_skinned:
            aspec = GeomVertexAnimationSpec()
            aspec.set_panda()
            vformat.set_animation(aspec)
            varray_blends = GeomVertexArrayFormat()
            varray_blends.add_column(InternalName.get_transform_blend(), 1, GeomEnums.NT_uint16, GeomEnums.C_index)

            vformat.add_array(varray_blends)
            vformat.add_array(varray_skin)
        reg_format = GeomVertexFormat.register_format(vformat)
        vdata = vdata.convert_to(reg_format)

        # Construct primitive
        primitiveid = geom_node.get_num_geoms()
        primitivemode = gltf_primitive.get('mode', 4)
        try:
            prim = self._PRIMITIVE_MODE_MAP[primitivemode](GeomEnums.UH_static)
        except KeyError:
            print(
                "Warning: primitive {} on mesh {} has an unsupported mode: {}"
                .format(primitiveid, geom_node.name, primitivemode)
            )
            return

        if 'indices' in gltf_primitive:
            index_acc = gltf_data['accessors'][gltf_primitive['indices']]
            prim.set_index_type(self._COMPONENT_TYPE_MAP[index_acc['componentType']])

            handle = prim.modify_vertices(index_acc['count']).modify_handle()
            handle.unclean_set_num_rows(index_acc['count'])

            buffview = gltf_data['bufferViews'][index_acc['bufferView']]
            buff = self.buffers[buffview['buffer']]
            start = buffview['byteOffset']
            end = start + index_acc['count'] * buffview.get('byteStride', 1) * prim.index_stride
            handle.copy_data_from(buff[start:end])
            handle = None

        # Assign a material
        matid = gltf_primitive.get('material', None)
        if matid is None:
            print(
                "Warning: mesh {} has a primitive with no material, using an empty RenderState"
                .format(geom_node.name)
            )
            mat = RenderState.make_empty()
        elif matid not in self.mat_states:
            print(
                "Warning: material with name {} has no associated mat state, using an empty RenderState"
                .format(matid)
            )
            mat = RenderState.make_empty()
        else:
            mat = self.mat_states[gltf_primitive['material']]
            self.mat_mesh_map[gltf_primitive['material']].append((geom_node.name, primitiveid))

        # Add this primitive back to the geom node
        #ss = StringStream()
        #vdata.write(ss)
        ###prim.write(ss, 2)
        #print(ss.data.decode('utf8'))
        geom = Geom(vdata)
        geom.add_primitive(prim)
        #geom.transform_vertices(self.csxform)
        geom_node.add_geom(geom, mat)

    def load_mesh(self, meshid, gltf_mesh, gltf_data):
        mesh_name = gltf_mesh.get('name', 'mesh'+str(meshid))
        node = self.meshes.get(meshid, GeomNode(mesh_name))

        # Clear any existing mesh data
        node.remove_all_geoms()

        # Load primitives
        for gltf_primitive in gltf_mesh['primitives']:
            self.load_primitive(node, gltf_primitive, gltf_data)

        # Save mesh
        self.meshes[meshid] = node

    def combine_mesh_skin(self, geom_node, skinid):
        def read_vert_data(gvd, column_name):
            gvr = GeomVertexReader(gvd, column_name)
            data = []
            while not gvr.is_at_end():
                data.append(LVecBase4(gvr.get_data4()))
            return data

        jvtmap = collections.OrderedDict(sorted(self.joint_map[skinid].items()))

        for geom in geom_node.modify_geoms():
            gvd = geom.modify_vertex_data()
            tbtable = TransformBlendTable()
            tdata = GeomVertexWriter(gvd, InternalName.get_transform_blend())

            jointdata = read_vert_data(gvd, InternalName.get_transform_index())
            weightdata = read_vert_data(gvd, InternalName.get_transform_weight())

            for joints, weights in zip(jointdata, weightdata):
                tblend = TransformBlend()
                for joint, weight in zip(joints, weights):
                    try:
                        jvt = jvtmap[joint]
                    except KeyError:
                        print(
                            "Could not find joint in jvtmap:\n\tjoint={}\n\tjvtmap={}"
                            .format(joint, jvtmap)
                        )
                        continue
                    tblend.add_transform(jvt, weight)
                tdata.add_data1i(tbtable.add_blend(tblend))
            tbtable.set_rows(SparseArray.lower_on(gvd.get_num_rows()))
            gvd.set_transform_blend_table(tbtable)

    def load_camera(self, camid, gltf_camera):
        camname = gltf_camera.get('name', 'cam'+str(camid))
        node = self.cameras.get(camid, Camera(camname))

        if gltf_camera['type'] == 'perspective':
            gltf_lens = gltf_camera['perspective']
            lens = PerspectiveLens()
            lens.set_fov(math.degrees(gltf_lens['yfov'] * gltf_lens['aspectRatio']), math.degrees(gltf_lens['yfov']))
            lens.set_near_far(gltf_lens['znear'], gltf_lens['zfar'])
            lens.set_view_vector((0, 0, -1), (0, 1, 0))
            node.set_lens(lens)

        self.cameras[camid] = node

    def load_light(self, lightid, gltf_light):
        node = self.lights.get(lightid, None)
        lightname = gltf_light.get('name', 'light'+str(lightid))

        ltype = gltf_light['type']
        # Construct a new light if needed
        if node is None:
            if ltype == 'point':
                node = PointLight(lightname)
            elif ltype == 'directional':
                node = DirectionalLight(lightname)
                node.set_direction((0, 0, -1))
            elif ltype == 'spot':
                node = Spotlight(lightname)
            else:
                print("Unsupported light type for light with name {}: {}".format(lightname, gltf_light['type']))
                node = PandaNode(lightname)

        # Update the light
        if ltype == 'unsupported':
            lightprops = {}
        else:
            lightprops = gltf_light[ltype]

        if ltype in ('point', 'directional', 'spot'):
            node.set_color(LColor(*lightprops['color'], w=1))

        if ltype in ('point', 'spot'):
            att = LPoint3(
                lightprops['constantAttenuation'],
                lightprops['linearAttenuation'],
                lightprops['quadraticAttenuation']
            )
            node.set_attenuation(att)

        self.lights[lightid] = node


def load_model(loader, file_path, **loader_kwargs):
    '''Load a glTF file from file_path and return a ModelRoot'''
    import tempfile
    import subprocess

    with tempfile.NamedTemporaryFile(suffix='.bam') as bamfile:
        try:
            subprocess.check_call(['gltf2bam', file_path, bamfile.name])
        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to convert glTF file")
        model_root = loader.load_model(bamfile.name, **loader_kwargs)

    return model_root


def main():
    import sys
    import json

    if len(sys.argv) < 2:
        print("Missing glTF srouce file argument")
        sys.exit(1)
    elif len(sys.argv) < 3 and not sys.argv[1].endswith('.gltf'):
        print("Missing bam destination file argument")
        sys.exit(1)

    infile = sys.argv[1]
    outfile = sys.argv[2] if len(sys.argv) > 2 else infile.replace('.gltf', '.bam')

    with open(infile) as gltf_file:
        gltf_data = json.load(gltf_file)

    dstfname = Filename.fromOsSpecific(outfile)

    indir = Filename(Filename.from_os_specific(infile).get_dirname())
    outdir = Filename(dstfname.get_dirname())

    get_model_path().prepend_directory(indir)
    get_model_path().prepend_directory(outdir)

    converter = Converter(indir=indir, outdir=outdir)
    converter.update(gltf_data, writing_bam=True)

    #converter.active_scene.ls()

    converter.active_scene.write_bam_file(dstfname)


if __name__ == '__main__':
    main()



import importlib
from distutils.version import StrictVersion as Version
import json
import os
import time

import bpy
try:
    import bpy_extras
    IS_IN_BLENDER = True
except ModuleNotFoundError:
    IS_IN_BLENDER = False

from . import blendergltf
from . import exporters
from . import filters
from . import pbr_utils
from . import extension_exporters as extensions

bl_info = {
    "name": "glTF format",
    "author": "Daniel Stokes and GitHub contributors",
    # When updating version, also update the generator string in blendergltf.py
    "version": (1, 2, 0),
    "blender": (2, 76, 0),
    "location": "File > Import-Export",
    "description": "Export glTF",
    "warning": "",
    "wiki_url": "https://github.com/Kupoman/blendergltf/blob/master/README.md"
                "",
    "tracker_url": "https://github.com/Kupoman/blendergltf/issues",
    "support": 'COMMUNITY',
    "category": "Import-Export"
}


__all__ = [
    blendergltf,
    exporters,
    extensions,
    filters,
    pbr_utils,
]


if "__IMPORTED__" in locals():
    importlib.reload(locals()['blendergltf'])
    importlib.reload(locals()['filters'])
    importlib.reload(locals()['extensions'])
    importlib.reload(locals()['pbr_utils'])

if IS_IN_BLENDER:
    GLTFOrientationHelper = bpy_extras.io_utils.orientation_helper_factory(
        "GLTFOrientationHelper", axis_forward='Z', axis_up='Y'
    )

    VERSION_ITEMS = (
        ('1.0', '1.0', ''),
        ('2.0', '2.0', ''),
    )

    PROFILE_ITEMS = (
        ('WEB', 'Web', 'Export shaders for WebGL 1.0 use (shader version 100)'),
        ('DESKTOP', 'Desktop', 'Export shaders for OpenGL 3.0 use (shader version 130)')
    )
    IMAGE_STORAGE_ITEMS = (
        ('EMBED', 'Embed', 'Embed image data into the glTF file'),
        ('REFERENCE', 'Reference', 'Use the same filepath that Blender uses for images'),
        ('COPY', 'Copy', 'Copy images to output directory and use a relative reference')
    )
    ANIM_EXPORT_ITEMS = (
        ('ACTIVE', 'Active Only', 'Export the active action per object'),
        ('ELIGIBLE', 'All Eligible', 'Export all actions that can be used by an object'),
    )

    class ExtPropertyGroup(bpy.types.PropertyGroup):
        name = bpy.props.StringProperty(name='Extension Name')
        enable = bpy.props.BoolProperty(
            name='Enable',
            description='Enable this extension',
            default=False
        )

    class ExportGLTF(bpy.types.Operator, bpy_extras.io_utils.ExportHelper, GLTFOrientationHelper):
        """Save a Khronos glTF File"""
        bl_idname = 'export_scene.gltf'
        bl_label = 'Export glTF'
        bl_options = {'PRESET'}

        filename_ext = ''
        filter_glob = bpy.props.StringProperty(
            default='*.gltf;*.glb',
            options={'HIDDEN'},
        )

        # Override filepath to simplify linting
        filepath = bpy.props.StringProperty(
            name='File Path',
            description='Filepath used for exporting the file',
            maxlen=1024,
            subtype='FILE_PATH'
        )

        check_extension = True

        ext_exporters = sorted(
            [exporter() for exporter in extensions.__all__],
            key=lambda ext: ext.ext_meta['name']
        )
        extension_props = bpy.props.CollectionProperty(
            name='Extensions',
            type=ExtPropertyGroup,
            description='Select extensions to enable'
        )
        ext_prop_to_exporter_map = {}
        for ext_exporter in ext_exporters:
            meta = ext_exporter.ext_meta
            if 'settings' in meta:
                name = 'settings_' + meta['name']
                prop_group = type(name, (bpy.types.PropertyGroup,), meta['settings'])
                bpy.utils.register_class(prop_group)
                value = bpy.props.PointerProperty(type=prop_group)
                locals()[name] = value

        # Dummy property to get icon with tooltip
        draft_prop = bpy.props.BoolProperty(
            name='',
            description='This extension is currently in a draft phase',
        )

        # blendergltf settings
        nodes_export_hidden = bpy.props.BoolProperty(
            name='Export Hidden Objects',
            description='Export nodes that are not set to visible',
            default=False
        )
        nodes_selected_only = bpy.props.BoolProperty(
            name='Selection Only',
            description='Only export nodes that are currently selected',
            default=False
        )
        materials_disable = bpy.props.BoolProperty(
            name='Disable Material Export',
            description='Export minimum default materials. Useful when using material extensions',
            default=False
        )
        meshes_apply_modifiers = bpy.props.BoolProperty(
            name='Apply Modifiers',
            description='Apply all modifiers to the output mesh data',
            default=True
        )
        meshes_vertex_color_alpha = bpy.props.BoolProperty(
            name='Export Vertex Color Alpha',
            description=(
                'Export vertex colors with 4 channels'
                ' (the value of the fourth channel is always 1.0)'
            ),
            default=False
        )
        meshes_interleave_vertex_data = bpy.props.BoolProperty(
            name='Interleave Vertex Data',
            description=(
                'Store data for each vertex contiguously'
                ' instead of each vertex property (e.g. position) contiguously'
            ),
            default=False
        )
        animations_object_export = bpy.props.EnumProperty(
            items=ANIM_EXPORT_ITEMS,
            name='Objects',
            default='ACTIVE'
        )
        animations_armature_export = bpy.props.EnumProperty(
            items=ANIM_EXPORT_ITEMS,
            name='Armatures',
            default='ELIGIBLE'
        )
        animations_shape_key_export = bpy.props.EnumProperty(
            items=ANIM_EXPORT_ITEMS,
            name='Shape Keys',
            default='ELIGIBLE'
        )
        images_data_storage = bpy.props.EnumProperty(
            items=IMAGE_STORAGE_ITEMS,
            name='Storage',
            default='COPY'
        )
        images_allow_srgb = bpy.props.BoolProperty(
            name='sRGB Texture Support',
            description='Use sRGB texture formats for sRGB textures',
            default=False
        )
        buffers_embed_data = bpy.props.BoolProperty(
            name='Embed Buffer Data',
            description='Embed buffer data into the glTF file',
            default=False
        )
        buffers_combine_data = bpy.props.BoolProperty(
            name='Combine Buffer Data',
            description='Combine all buffers into a single buffer',
            default=True
        )
        asset_copyright = bpy.props.StringProperty(
            name='Copyright',
            description='Copyright string to include in output file'
        )
        asset_version = bpy.props.EnumProperty(
            items=VERSION_ITEMS,
            name='Version',
            default='2.0'
        )
        asset_profile = bpy.props.EnumProperty(
            items=PROFILE_ITEMS,
            name='Profile',
            default='WEB'
        )
        gltf_export_binary = bpy.props.BoolProperty(
            name='Export as binary',
            description='Export to the binary glTF file format (.glb)',
            default=False
        )
        pretty_print = bpy.props.BoolProperty(
            name='Pretty-print / indent JSON',
            description='Export JSON with indentation and a newline',
            default=True
        )
        blocks_prune_unused = bpy.props.BoolProperty(
            name='Prune Unused Resources',
            description='Do not export any data-blocks that have no users or references',
            default=True
        )
        enable_actions = bpy.props.BoolProperty(
            name='Actions',
            description='Enable the export of actions',
            default=True
        )
        enable_cameras = bpy.props.BoolProperty(
            name='Cameras',
            description='Enable the export of cameras',
            default=True
        )
        enable_lamps = bpy.props.BoolProperty(
            name='Lamps',
            description='Enable the export of lamps',
            default=True
        )
        enable_materials = bpy.props.BoolProperty(
            name='Materials',
            description='Enable the export of materials',
            default=True
        )
        enable_meshes = bpy.props.BoolProperty(
            name='Meshes',
            description='Enable the export of meshes',
            default=True
        )
        enable_textures = bpy.props.BoolProperty(
            name='Textures',
            description='Enable the export of textures',
            default=True
        )

        def update_extensions(self):
            self.ext_prop_to_exporter_map = {
                ext.ext_meta['name']: ext for ext in self.ext_exporters
            }

            for exporter in self.ext_exporters:
                exporter.ext_meta['enable'] = False
            for prop in self.extension_props:
                exporter = self.ext_prop_to_exporter_map[prop.name]
                exporter.ext_meta['enable'] = prop.enable

            self.extension_props.clear()
            for exporter in self.ext_exporters:
                prop = self.extension_props.add()
                prop.name = exporter.ext_meta['name']
                prop.enable = exporter.ext_meta['enable']

                if exporter.ext_meta['name'] == 'KHR_technique_webgl':
                    prop.enable = Version(self.asset_version) < Version('2.0')

        def invoke(self, context, event):
            self.update_extensions()
            return super().invoke(context, event)

        def check(self, context):
            redraw = False

            if self.gltf_export_binary and self.filepath.endswith('.gltf'):
                self.filepath = self.filepath[:-4] + 'glb'
                redraw = True
            elif not self.gltf_export_binary and self.filepath.endswith('.glb'):
                self.filepath = self.filepath[:-3] + 'gltf'
                redraw = True

            if (
                    self.gltf_export_binary
                    and self.buffers_embed_data
                    and not self.buffers_combine_data
            ):
                self.buffers_combine_data = True
                redraw = True

            self.filename_ext = '.glb' if self.gltf_export_binary else '.gltf'
            redraw = redraw or super().check(context)

            return redraw

        def draw(self, context):
            self.update_extensions()
            layout = self.layout

            col = layout.box().column(align=True)
            col.label('Enable:')
            row = col.row(align=True)
            row.prop(self, 'enable_actions', toggle=True)
            row.prop(self, 'enable_cameras', toggle=True)
            row.prop(self, 'enable_lamps', toggle=True)
            row = col.row(align=True)
            row.prop(self, 'enable_materials', toggle=True)
            row.prop(self, 'enable_meshes', toggle=True)
            row.prop(self, 'enable_textures', toggle=True)

            col = layout.box().column()
            col.label('Axis Conversion:', icon='MANIPUL')
            col.prop(self, 'axis_up')
            col.prop(self, 'axis_forward')

            col = layout.box().column()
            col.label('Nodes:', icon='OBJECT_DATA')
            col.prop(self, 'nodes_export_hidden')
            col.prop(self, 'nodes_selected_only')

            col = layout.box().column()
            col.label('Meshes:', icon='MESH_DATA')
            col.prop(self, 'meshes_apply_modifiers')
            col.prop(self, 'meshes_interleave_vertex_data')
            col.prop(self, 'meshes_vertex_color_alpha')

            col = layout.box().column()
            col.label('Materials:', icon='MATERIAL_DATA')
            col.prop(self, 'materials_disable')
            if Version(self.asset_version) < Version('2.0'):
                material_settings = getattr(self, 'settings_KHR_technique_webgl')
                col.prop(material_settings, 'embed_shaders')

            col = layout.box().column()
            col.label('Animations:', icon='ACTION')
            col.prop(self, 'animations_armature_export')
            col.prop(self, 'animations_object_export')
            col.prop(self, 'animations_shape_key_export')

            col = layout.box().column()
            col.label('Images:', icon='IMAGE_DATA')
            col.prop(self, 'images_data_storage')
            if Version(self.asset_version) < Version('2.0'):
                col.prop(self, 'images_allow_srgb')

            col = layout.box().column()
            col.label('Buffers:', icon='SORTALPHA')
            col.prop(self, 'buffers_embed_data')

            col = col.column()
            col.enabled = not self.gltf_export_binary or not self.buffers_embed_data
            prop = col.prop(self, 'buffers_combine_data')

            col = layout.box().column()
            col.label('Extensions:', icon='PLUGIN')
            extension_filter = set()

            # Disable KHR_technique_webgl for all glTF versions
            extension_filter.add('KHR_technique_webgl')
            for i in range(len(self.extension_props)):
                prop = self.extension_props[i]
                extension_exporter = self.ext_prop_to_exporter_map[prop.name]

                if extension_exporter.ext_meta['name'] in extension_filter:
                    continue

                row = col.row()
                row.prop(prop, 'enable', text=prop.name)
                if extension_exporter.ext_meta.get('isDraft', False):
                    row.prop(self, 'draft_prop', icon='ERROR', emboss=False)
                info_op = row.operator('wm.url_open', icon='INFO', emboss=False)
                info_op.url = extension_exporter.ext_meta.get('url', '')

                if prop.enable:
                    settings = getattr(self, 'settings_' + prop.name, None)
                    if settings:
                        box = col.box()
                        if hasattr(extension_exporter, 'draw_settings'):
                            extension_exporter.draw_settings(box, settings, context)
                        else:
                            setting_props = [
                                name for name in dir(settings)
                                if not name.startswith('_')
                                and name not in ('bl_rna', 'name', 'rna_type')
                            ]
                            for setting_prop in setting_props:
                                box.prop(settings, setting_prop)
                        if i < len(self.extension_props) - 1:
                            col.separator()
                            col.separator()

            col = layout.box().column()
            col.label('Output:', icon='SCRIPTWIN')
            col.prop(self, 'asset_copyright')
            col.prop(self, 'asset_version')
            col.prop(self, 'gltf_export_binary')
            if Version(self.asset_version) < Version('2.0'):
                col.prop(self, 'asset_profile')
            col.prop(self, 'pretty_print')
            col.prop(self, 'blocks_prune_unused')

        def execute(self, _):
            # Copy properties to settings
            settings = self.as_keywords(ignore=(
                "filter_glob",
                "axis_up",
                "axis_forward",
            ))

            # Set the output directory based on the supplied file path
            settings['gltf_output_dir'] = os.path.dirname(self.filepath)

            # Set the output name
            settings['gltf_name'] = os.path.splitext(os.path.basename(self.filepath))[0]

            # Calculate a global transform matrix to apply to a root node
            settings['nodes_global_matrix'] = bpy_extras.io_utils.axis_conversion(
                to_forward=self.axis_forward,
                to_up=self.axis_up
            ).to_4x4()

            # filter data according to settings
            data = {
                'actions': list(bpy.data.actions) if self.enable_actions else [],
                'cameras': list(bpy.data.cameras) if self.enable_cameras else [],
                'lamps': list(bpy.data.lamps) if self.enable_lamps else [],
                'images': list(bpy.data.images) if self.enable_textures else [],
                'materials': list(bpy.data.materials) if self.enable_materials else [],
                'meshes': list(bpy.data.meshes) if self.enable_meshes else [],
                'objects': list(bpy.data.objects),
                'scenes': list(bpy.data.scenes),
                'textures': list(bpy.data.textures) if self.enable_textures else [],
            }

            # Remove objects that point to disabled data
            if not self.enable_cameras:
                data['objects'] = [
                    obj for obj in data['objects']
                    if not isinstance(obj.data, bpy.types.Camera)
                ]
            if not self.enable_lamps:
                data['objects'] = [
                    obj for obj in data['objects']
                    if not isinstance(obj.data, bpy.types.Lamp)
                ]
            if not self.enable_meshes:
                data['objects'] = [
                    obj for obj in data['objects']
                    if not isinstance(obj.data, bpy.types.Mesh)
                ]

            if not settings['nodes_export_hidden']:
                data = filters.visible_only(data)

            if settings['nodes_selected_only']:
                data = filters.selected_only(data)

            if settings['blocks_prune_unused']:
                data = filters.used_only(data)

            for ext_exporter in self.ext_exporters:
                ext_exporter.settings = getattr(
                    self,
                    'settings_' + ext_exporter.ext_meta['name'],
                    None
                )

            def is_builtin_mat_ext(prop_name):
                if Version(self.asset_version) < Version('2.0'):
                    return prop_name == 'KHR_technique_webgl'
                return False

            settings['extension_exporters'] = [
                self.ext_prop_to_exporter_map[prop.name]
                for prop in self.extension_props
                if prop.enable and not (self.materials_disable and is_builtin_mat_ext(prop.name))
            ]

            start_time = time.perf_counter()
            gltf = blendergltf.export_gltf(data, settings)
            end_time = time.perf_counter()
            print('Export took {:.4} seconds'.format(end_time - start_time))

            if self.gltf_export_binary:
                with open(self.filepath, 'wb') as fout:
                    fout.write(gltf)
            else:
                with open(self.filepath, 'w') as fout:
                    # Figure out indentation
                    indent = 4 if self.pretty_print else None

                    # Dump the JSON
                    json.dump(gltf, fout, indent=indent, sort_keys=True, check_circular=False)

                    if self.pretty_print:
                        # Write a newline to the end of the file
                        fout.write('\n')
            return {'FINISHED'}

    def menu_func_export(self, _):
        self.layout.operator(ExportGLTF.bl_idname, text="glTF (.gltf)")

    def register():
        bpy.utils.register_class(ExtPropertyGroup)
        bpy.utils.register_class(ExportGLTF)
        bpy.utils.register_class(pbr_utils.PbrExportPanel)
        bpy.utils.register_class(pbr_utils.PbrSettings)

        bpy.types.Material.pbr_export_settings = bpy.props.PointerProperty(
            type=pbr_utils.PbrSettings
        )

        bpy.types.INFO_MT_file_export.append(menu_func_export)

    def unregister():
        bpy.types.INFO_MT_file_export.remove(menu_func_export)

        del bpy.types.Material.pbr_export_settings

        bpy.utils.unregister_class(pbr_utils.PbrSettings)
        bpy.utils.unregister_class(pbr_utils.PbrExportPanel)
        bpy.utils.unregister_class(ExportGLTF)
        bpy.utils.unregister_class(ExtPropertyGroup)


__IMPORTED__ = True



import os
import sys

from direct.actor.Actor import Actor
from direct.showbase.ShowBase import ShowBase
import panda3d.core as p3d

import gltf

p3d.load_prc_file_data(
    __file__,
    'window-size 1024 768\n'
)


class App(ShowBase):
    def __init__(self):
        if len(sys.argv) < 2:
            print("Missing input file")
            sys.exit(1)

        super().__init__()

        gltf.patch_loader(self.loader)

        infile = p3d.Filename.from_os_specific(os.path.abspath(sys.argv[1]))
        p3d.get_model_path().prepend_directory(infile.get_dirname())

        self.model_root = self.loader.load_model(infile)

        self.accept('escape', sys.exit)
        self.accept('q', sys.exit)

        viewer_dir = os.path.dirname(__file__)
        vertfname = os.path.join(viewer_dir, 'simplepbr.vert')
        fragfname = os.path.join(viewer_dir, 'simplepbr.frag')
        pbrshader = p3d.Shader.load(p3d.Shader.SL_GLSL, vertex=vertfname, fragment=fragfname)
        self.render.set_shader(pbrshader)

        self.light = self.render.attach_new_node(p3d.PointLight('light'))
        self.light.set_pos(-5, 5, 5)
        self.render.set_light(self.light)

        self.cam.set_pos(-10, 10, 10)

        if self.model_root.find('**/+Character'):
            self.actor = Actor(self.model_root)
            self.actor.reparent_to(self.render)
            self.actor.loop('anim0')
            self.cam.look_at(self.actor)
        else:
            self.model_root.reparent_to(self.render)
            self.cam.look_at(self.model_root)

def main():
    App().run()

if __name__ == '__main__':
    main()



from setuptools import setup

def readme():
    with open('README.md') as readme_file:
        return readme_file.read()

setup(
    name='panda3d-gltf',
    version='0.1.0',
    description='glTF utilities for Panda3D',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
    ],
    keywords='panda3d gltf',
    packages=['gltf'],
    install_requires=[
        'panda3d',
    ],
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'pytest',
        'pylint',
        'pytest-pylint',
    ],
    entry_points={
        'console_scripts': [
            'gltf2bam=gltf.converter:main'
        ],
        'gui_scripts': [
            'gltf-viewer=gltf.viewer:main'
        ],
    },
)



import os
import subprocess


def run_blender(args, blenderdir=''):
    os.path.join(blenderdir, 'blender')
    subprocess.check_call(['blender', '--background'] + args, stdout=subprocess.DEVNULL)


def run_blender_script(script, args, blenderdir=''):
    run_blender(
        [
            '-P', script,
            '--',
        ] + args,
        blenderdir=blenderdir
    )



#!/bin/env python
import bpy


def visible_only(bpy_data):
    """
    Filter out all the invisible objects
    """

    def visible(obj):
        return any(obj.is_visible(s) for s in bpy_data['scenes'])

    bpy_data['objects'] = [obj for obj in bpy_data['objects'] if visible(obj)]
    return bpy_data


def selected_only(bpy_data):
    """
    Filter out all the objects that are not selected
    """

    def selected_in_subtree(parent_obj):
        return parent_obj.select or any(selected_in_subtree(child) for child in parent_obj.children)

    bpy_data['objects'] = [obj for obj in bpy_data['objects'] if selected_in_subtree(obj)]
    return bpy_data


def used_only(bpy_data):
    """
    Filter out the data blocks not in use by the objects in the list
    """

    # Clear flags
    for data_list in bpy_data.values():
        for data in data_list:
            data.tag = False

    # Always mark actions as used
    for action in bpy_data['actions']:
        action.tag = True

    # Need a function to handle some recursion for dupli-groups
    def tag_object(obj):
        obj.tag = True

        if obj.dupli_group:
            for dupli_obj in obj.dupli_group.objects:
                tag_object(dupli_obj)

        for scene in obj.users_scene:
            scene.tag = True

        if obj.data:
            obj.data.tag = True

        materials = []
        if obj.data and hasattr(obj.data, 'materials'):
            materials = [material for material in obj.data.materials.values() if material]

        for material in materials:
            material.tag = True

            textures = [
                slot.texture for slot in material.texture_slots.values()
                if slot is not None
                and slot.use
                and isinstance(slot.texture, bpy.types.ImageTexture)
            ]
            for texture in textures:
                texture.tag = True

                if texture.image:
                    texture.image.tag = True

    # Now let's go through the objects and start tagging
    for obj in bpy_data['objects']:
        tag_object(obj)

    pruned_data = {
        key: [data for data in value if data.tag] for key, value in bpy_data.items()
    }

    return pruned_data



import base64
import os
import re

import bpy
import gpu

LAMP_TYPES = [
    gpu.GPU_DYNAMIC_LAMP_DYNVEC,
    gpu.GPU_DYNAMIC_LAMP_DYNCO,
    gpu.GPU_DYNAMIC_LAMP_DYNIMAT,
    gpu.GPU_DYNAMIC_LAMP_DYNPERSMAT,
    gpu.GPU_DYNAMIC_LAMP_DYNENERGY,
    gpu.GPU_DYNAMIC_LAMP_DYNENERGY,
    gpu.GPU_DYNAMIC_LAMP_DYNCOL,
    gpu.GPU_DYNAMIC_LAMP_DISTANCE,
    gpu.GPU_DYNAMIC_LAMP_ATT1,
    gpu.GPU_DYNAMIC_LAMP_ATT2,
    gpu.GPU_DYNAMIC_LAMP_SPOTSIZE,
    gpu.GPU_DYNAMIC_LAMP_SPOTBLEND,
]

MIST_TYPES = [
    gpu.GPU_DYNAMIC_MIST_ENABLE,
    gpu.GPU_DYNAMIC_MIST_START,
    gpu.GPU_DYNAMIC_MIST_DISTANCE,
    gpu.GPU_DYNAMIC_MIST_INTENSITY,
    gpu.GPU_DYNAMIC_MIST_TYPE,
    gpu.GPU_DYNAMIC_MIST_COLOR,
]

WORLD_TYPES = [
    gpu.GPU_DYNAMIC_HORIZON_COLOR,
    gpu.GPU_DYNAMIC_AMBIENT_COLOR,
]

MATERIAL_TYPES = [
    gpu.GPU_DYNAMIC_MAT_DIFFRGB,
    gpu.GPU_DYNAMIC_MAT_REF,
    gpu.GPU_DYNAMIC_MAT_SPECRGB,
    gpu.GPU_DYNAMIC_MAT_SPEC,
    gpu.GPU_DYNAMIC_MAT_HARD,
    gpu.GPU_DYNAMIC_MAT_EMIT,
    gpu.GPU_DYNAMIC_MAT_AMB,
    gpu.GPU_DYNAMIC_MAT_ALPHA,
]

TYPE_TO_NAME = {
    gpu.GPU_DYNAMIC_OBJECT_VIEWMAT: 'view_mat',
    gpu.GPU_DYNAMIC_OBJECT_MAT: 'model_mat',
    gpu.GPU_DYNAMIC_OBJECT_VIEWIMAT: 'inv_view_mat',
    gpu.GPU_DYNAMIC_OBJECT_IMAT: 'inv_model_mat',
    gpu.GPU_DYNAMIC_OBJECT_COLOR: 'color',
    gpu.GPU_DYNAMIC_OBJECT_AUTOBUMPSCALE: 'auto_bump_scale',

    gpu.GPU_DYNAMIC_MIST_ENABLE: 'use_mist',
    gpu.GPU_DYNAMIC_MIST_START: 'start',
    gpu.GPU_DYNAMIC_MIST_DISTANCE: 'depth',
    gpu.GPU_DYNAMIC_MIST_INTENSITY: 'intensity',
    gpu.GPU_DYNAMIC_MIST_TYPE: 'falloff',
    gpu.GPU_DYNAMIC_MIST_COLOR: 'color',

    gpu.GPU_DYNAMIC_HORIZON_COLOR: 'horizon_color',
    gpu.GPU_DYNAMIC_AMBIENT_COLOR: 'ambient_color',

    gpu.GPU_DYNAMIC_LAMP_DYNVEC: 'dynvec',
    gpu.GPU_DYNAMIC_LAMP_DYNCO: 'dynco',
    gpu.GPU_DYNAMIC_LAMP_DYNIMAT: 'dynimat',
    gpu.GPU_DYNAMIC_LAMP_DYNPERSMAT: 'dynpersmat',
    gpu.GPU_DYNAMIC_LAMP_DYNENERGY: 'energy',
    gpu.GPU_DYNAMIC_LAMP_DYNCOL: 'color',
    gpu.GPU_DYNAMIC_LAMP_DISTANCE: 'distance',
    gpu.GPU_DYNAMIC_LAMP_ATT1: 'linear_attenuation',
    gpu.GPU_DYNAMIC_LAMP_ATT2: 'quadratic_attenuation',
    gpu.GPU_DYNAMIC_LAMP_SPOTSIZE: 'spot_size',
    gpu.GPU_DYNAMIC_LAMP_SPOTBLEND: 'spot_blend',

    gpu.GPU_DYNAMIC_MAT_DIFFRGB: 'diffuse_color',
    gpu.GPU_DYNAMIC_MAT_REF: 'diffuse_intensity',
    gpu.GPU_DYNAMIC_MAT_SPECRGB: 'specular_color',
    gpu.GPU_DYNAMIC_MAT_SPEC: 'specular_intensity',
    gpu.GPU_DYNAMIC_MAT_HARD: 'specular_hardness',
    gpu.GPU_DYNAMIC_MAT_EMIT: 'emit',
    gpu.GPU_DYNAMIC_MAT_AMB: 'ambient',
    gpu.GPU_DYNAMIC_MAT_ALPHA: 'alpha',
}

TYPE_TO_SEMANTIC = {
    gpu.GPU_DYNAMIC_LAMP_DYNVEC: 'BL_DYNVEC',
    gpu.GPU_DYNAMIC_LAMP_DYNCO: 'MODELVIEW',  # dynco gets extracted from the matrix
    gpu.GPU_DYNAMIC_LAMP_DYNIMAT: 'BL_DYNIMAT',
    gpu.GPU_DYNAMIC_LAMP_DYNPERSMAT: 'BL_DYNPERSMAT',
    gpu.CD_ORCO: 'POSITION',
    gpu.CD_MTFACE: 'TEXCOORD_0',
    -1: 'NORMAL'        # Hack until the gpu module has something for normals
}

DATATYPE_TO_CONVERTER = {
    gpu.GPU_DATA_1I: lambda x: x,
    gpu.GPU_DATA_1F: lambda x: x,
    gpu.GPU_DATA_2F: list,
    gpu.GPU_DATA_3F: list,
    gpu.GPU_DATA_4F: list,
}

DATATYPE_TO_GLTF_TYPE = {
    gpu.GPU_DATA_1I: 5124,  # INT
    gpu.GPU_DATA_1F: 5126,  # FLOAT
    gpu.GPU_DATA_2F: 35664,  # FLOAT_VEC2
    gpu.GPU_DATA_3F: 35665,  # FLOAT_VEC3
    gpu.GPU_DATA_4F: 35666,  # FLOAT_VEC4
    gpu.GPU_DATA_9F: 35675,  # FLOAT_MAT3
    gpu.GPU_DATA_16F: 35676,  # FLOAT_MAT4
}


def vs_to_130(data):
    data['attributes'].append({
        'varname': 'bl_Vertex',
        'type': gpu.CD_ORCO,
        'datatype': gpu.GPU_DATA_4F
    })
    data['attributes'].append({
        'varname': 'bl_Normal',
        'type': -1,
        'datatype': gpu.GPU_DATA_3F
    })
    data['uniforms'].append({
        'varname': 'bl_ModelViewMatrix',
        'type': 0,
        'datatype': gpu.GPU_DATA_16F,
    })
    data['uniforms'].append({
        'varname': 'bl_ProjectionMatrix',
        'type': 0,
        'datatype': gpu.GPU_DATA_16F,
    })
    data['uniforms'].append({
        'varname': 'bl_NormalMatrix',
        'type': 0,
        'datatype': gpu.GPU_DATA_9F,
    })

    src = '#version 130\n'
    src += 'in vec4 bl_Vertex;\n'
    src += 'in vec3 bl_Normal;\n'
    src += 'uniform mat4 bl_ModelViewMatrix;\n'
    src += 'uniform mat4 bl_ProjectionMatrix;\n'
    src += 'uniform mat3 bl_NormalMatrix;\n'

    src += data['vertex']

    src = re.sub(r'#ifdef USE_OPENSUBDIV([^#]*)#endif', '', src)
    src = re.sub(r'#ifndef USE_OPENSUBDIV([^#]*)#endif', r'\1', src)
    src = re.sub(r'#ifdef CLIP_WORKAROUND(.*?)#endif', '', src, 0, re.DOTALL)
    src = re.sub(r'\bvarying\b', 'out', src)
    src = re.sub(r'\bgl_(?!Position)(.*?)\b', r'bl_\1', src)

    data['vertex'] = src


def fs_to_130(data):
    src = '#version 130\n'
    src += 'out vec4 frag_color;\n'
    src += 'uniform mat4 bl_ProjectionMatrix;\n'
    src += 'uniform mat4 bl_ModelViewMatrix;\n'
    src += 'uniform mat4 bl_ModelViewMatrixInverse;\n'
    src += 'uniform mat3 bl_NormalMatrix;\n'
    src += 'uniform mat4 bl_ProjectionMatrixInverse;\n'

    src += data['fragment']

    src = re.sub(r'\bvarying\b', 'in', src)
    src = re.sub(r'\bgl_FragColor\b', 'frag_color', src)
    src = re.sub(r'\bgl_(?!FrontFacing)(.*?)\b', r'bl_\1', src)

    # Cannot support node_bsdf functions without resolving use of gl_Light
    src = re.sub(r'void node_((bsdf)|(subsurface))_.*?^}', '', src, 0, re.DOTALL | re.MULTILINE)

    # Need to gather light data from more general uniforms
    light_count = 0
    light_map = {}
    decl_start_str = 'void main()\n{\n'
    for uniform in data['uniforms']:
        if uniform['type'] == gpu.GPU_DYNAMIC_LAMP_DYNCO:
            lamp_name = uniform['lamp'].name
            if lamp_name not in light_map:
                light_map[lamp_name] = light_count
                light_count += 1
            light_index = light_map[lamp_name]
            varname = 'light{}_transform'.format(light_index)
            uniform['datatype'] = gpu.GPU_DATA_16F
            src = src.replace(
                'uniform vec3 {};'.format(uniform['varname']),
                'uniform mat4 {};'.format(varname)
            )
            var_decl_start = src.find(decl_start_str) + len(decl_start_str)
            decl_str = '\tvec3 {} = {}[3].xyz;\n'.format(uniform['varname'], varname)
            src = src[:var_decl_start] + decl_str + src[var_decl_start:]
            uniform['varname'] = varname

    data['fragment'] = src.replace('\r\r\n', '')


def vs_to_web(data):
    src = data['vertex']

    precision_block = '\n'
    for data_type in ('float', 'int'):
        precision_block += 'precision mediump {};\n'.format(data_type)

    src = src.replace('#version 130', '#version 100\n' + precision_block)
    src = re.sub(r'\bin\b', 'attribute', src)
    src = re.sub(r'\bout\b', 'varying', src)

    data['vertex'] = src


def fs_to_web(data):
    src = data['fragment']

    precision_block = '\n'
    for data_type in ('float', 'int'):
        precision_block += 'precision mediump {};\n'.format(data_type)

    header = '#version 100\n'
    header += '#extension GL_OES_standard_derivatives: enable\n'
    header += precision_block
    src = src.replace('#version 130', header)
    src = re.sub(r'\bin\b', 'varying', src)
    src = src.replace('out vec4 frag_color;\n', '')
    src = re.sub(r'\bfrag_color\b', 'gl_FragColor', src)

    # TODO: This should be fixed in Blender
    src = src.replace('blend = (normalize(vec).z + 1)', 'blend = (normalize(vec).z + 1.0)')

    # TODO: This likely breaks shadows
    src = src.replace('sampler2DShadow', 'sampler2D')
    src = src.replace('shadow2DProj', 'texture2DProj')

    data['fragment'] = src


def to_130(data):
    vs_to_130(data)
    fs_to_130(data)


def to_web(data):
    to_130(data)
    vs_to_web(data)
    fs_to_web(data)


class KhrTechniqueWebgl:
    ext_meta = {
        'name': 'KHR_technique_webgl',
        'url': (
            'https://github.com/KhronosGroup/glTF/tree/master/extensions/'
            'Khronos/KHR_technique_webgl'
        ),
        'isDraft': True,
        'settings': {
            'embed_shaders': bpy.props.BoolProperty(
                name='Embed Shader Data',
                description='Embed shader data into the glTF file instead of a separate file',
                default=False
            )
        }
    }
    settings = None

    def export_material(self, state, material):
        shader_data = gpu.export_shader(bpy.context.scene, material)
        if state['settings']['asset_profile'] == 'DESKTOP':
            to_130(shader_data)
        else:
            to_web(shader_data)

        if self.settings.embed_shaders is True:
            fs_bytes = shader_data['fragment'].encode()
            fs_uri = 'data:text/plain;base64,' + base64.b64encode(fs_bytes).decode('ascii')
            vs_bytes = shader_data['vertex'].encode()
            vs_uri = 'data:text/plain;base64,' + base64.b64encode(vs_bytes).decode('ascii')
        else:
            names = [
                bpy.path.clean_name(name) + '.glsl'
                for name in (material.name+'VS', material.name+'FS')
            ]
            data = (shader_data['vertex'], shader_data['fragment'])
            for name, data in zip(names, data):
                filename = os.path.join(state['settings']['gltf_output_dir'], name)
                with open(filename, 'w') as fout:
                    fout.write(data)
            vs_uri, fs_uri = names

        state['output']['shaders'].append({
            'type': 35632,
            'uri': fs_uri,
            'name': material.name + 'FS',
        })
        state['output']['shaders'].append({
            'type': 35633,
            'uri': vs_uri,
            'name': material.name + 'VS',
        })

        # Handle programs
        state['output']['programs'].append({
            'attributes': [a['varname'] for a in shader_data['attributes']],
            'fragmentShader': 'shaders_{}FS'.format(material.name),
            'vertexShader': 'shaders_{}VS'.format(material.name),
            'name': material.name,
        })

        # Handle parameters/values
        values = {}
        parameters = {}
        for attribute in shader_data['attributes']:
            name = attribute['varname']
            semantic = TYPE_TO_SEMANTIC[attribute['type']]
            _type = DATATYPE_TO_GLTF_TYPE[attribute['datatype']]
            parameters[name] = {'semantic': semantic, 'type': _type}

        for uniform in shader_data['uniforms']:
            valname = TYPE_TO_NAME.get(uniform['type'], uniform['varname'])
            rnaname = valname
            semantic = None
            node = None
            value = None

            if uniform['varname'] == 'bl_ModelViewMatrix':
                semantic = 'MODELVIEW'
            elif uniform['varname'] == 'bl_ProjectionMatrix':
                semantic = 'PROJECTION'
            elif uniform['varname'] == 'bl_NormalMatrix':
                semantic = 'MODELVIEWINVERSETRANSPOSE'
            else:
                if uniform['type'] in LAMP_TYPES:
                    node = uniform['lamp'].name
                    valname = node + '_' + valname
                    semantic = TYPE_TO_SEMANTIC.get(uniform['type'], None)
                    if not semantic:
                        lamp_obj = bpy.data.objects[node]
                        value = getattr(lamp_obj.data, rnaname)
                elif uniform['type'] in MIST_TYPES:
                    valname = 'mist_' + valname
                    mist_settings = bpy.context.scene.world.mist_settings
                    if valname == 'mist_color':
                        value = bpy.context.scene.world.horizon_color
                    else:
                        value = getattr(mist_settings, rnaname)

                    if valname == 'mist_falloff':
                        if value == 'QUADRATIC':
                            value = 0.0
                        elif value == 'LINEAR':
                            value = 1.0
                        else:
                            value = 2.0
                elif uniform['type'] in WORLD_TYPES:
                    world = bpy.context.scene.world
                    value = getattr(world, rnaname)
                elif uniform['type'] in MATERIAL_TYPES:
                    converter = DATATYPE_TO_CONVERTER[uniform['datatype']]
                    value = converter(getattr(material, rnaname))
                    values[valname] = value
                elif uniform['type'] == gpu.GPU_DYNAMIC_SAMPLER_2DIMAGE:
                    texture_slots = [
                        slot for slot in material.texture_slots
                        if slot and slot.texture.type == 'IMAGE'
                    ]
                    for slot in texture_slots:
                        if slot.texture.image.name == uniform['image'].name:
                            value = 'texture_' + slot.texture.name
                            values[uniform['varname']] = value
                else:
                    print('Unconverted uniform:', uniform)

            parameter = {}
            if semantic:
                parameter['semantic'] = semantic
                if node:
                    parameter['node'] = 'node_' + node
            elif value:
                parameter['value'] = DATATYPE_TO_CONVERTER[uniform['datatype']](value)
            else:
                parameter['value'] = None

            if uniform['type'] == gpu.GPU_DYNAMIC_SAMPLER_2DIMAGE:
                parameter['type'] = 35678  # SAMPLER_2D
            else:
                parameter['type'] = DATATYPE_TO_GLTF_TYPE[uniform['datatype']]
            parameters[valname] = parameter
            uniform['valname'] = valname

        # Handle techniques
        tech_name = 'techniques_' + material.name
        state['output']['techniques'].append({
            'parameters': parameters,
            'program': 'programs_' + material.name,
            'attributes': {a['varname']: a['varname'] for a in shader_data['attributes']},
            'uniforms': {u['varname']: u['valname'] for u in shader_data['uniforms']},
            'name': material.name,
        })

        return {'technique': tech_name, 'values': values, 'name': material.name}

    def export(self, state):
        state['output']['techniques'] = []
        state['output']['shaders'] = []
        state['output']['programs'] = []
        state['output']['materials'] = [
            self.export_material(state, bl_mat) for bl_mat in state['input']['materials']
        ]



from distutils.version import StrictVersion as Version

from .base import BaseExporter

from .common import (
    Reference,
    SimpleID,
    get_bone_name,
)


class NodeExporter(BaseExporter):
    gltf_key = 'nodes'
    blender_key = 'objects'

    @classmethod
    def export(cls, state, blender_data):
        node = {
            'name': blender_data.name,
        }

        obj_children = [
            child for child in blender_data.children
            if child in state['input']['objects'] and not child.parent_bone
        ]
        if obj_children:
            node['children'] = []
        for i, child in enumerate(obj_children):
            node['children'].append(Reference('objects', child.name, node['children'], i))
            state['references'].append(node['children'][-1])

        extras = cls.get_custom_properties(blender_data)
        extras.update({
            prop.name: prop.value for prop in blender_data.game.properties.values()
        })
        if extras:
            node['extras'] = extras

        decompose = state['decompose_fn']

        if blender_data.parent and blender_data.parent_bone:
            parent_bone = blender_data.parent.data.bones[blender_data.parent_bone]
            bone_name = get_bone_name(parent_bone)
            if bone_name not in state['bone_children']:
                state['bone_children'][bone_name] = []
            state['bone_children'][bone_name].append(blender_data.name)

        if blender_data.type == 'MESH':
            decompose = state['decompose_mesh_fn']
            mesh = state['mod_meshes_obj'].get(blender_data.name, blender_data.data)
            mesh_name = mesh.name
            mesh = state['mod_meshes'].get(mesh.name, mesh)
            if state['version'] < Version('2.0'):
                node['meshes'] = []
                node['meshes'].append(Reference('meshes', mesh_name, node['meshes'], 0))
                state['references'].append(node['meshes'][0])
            else:
                node['mesh'] = Reference('meshes', mesh_name, node, 'mesh')
                state['references'].append(node['mesh'])
            armature = blender_data.find_armature()
            if armature:
                state['input']['skins'].append(blender_data)
                state['skinned_meshes'].add(mesh_name)
                node['skin'] = Reference('skins', blender_data.name, node, 'skin')
                state['references'].append(node['skin'])
                if state['version'] < Version('2.0'):
                    bone_names = [
                        get_bone_name(b) for b in armature.data.bones if b.parent is None
                    ]
                    node['skeletons'] = []
                    node['skeletons'].extend([
                        Reference('objects', bone, node['skeletons'], i)
                        for i, bone in enumerate(bone_names)
                    ])
                    for ref in node['skeletons']:
                        state['references'].append(ref)
        elif blender_data.type == 'CAMERA':
            node['camera'] = Reference('cameras', blender_data.data.name, node, 'camera')
            state['references'].append(node['camera'])
        elif blender_data.type == 'EMPTY' and blender_data.dupli_group is not None:
            node['children'] = node.get('children', [])
            node['children'].append(cls.export_dupli_group(state, blender_data.dupli_group))
            node['children'][-1].source = node['children']
            node['children'][-1].prop = len(node['children']) - 1
        elif blender_data.type == 'ARMATURE':
            decompose = state['decompose_mesh_fn']
            for i, bone in enumerate(blender_data.data.bones):
                state['input']['bones'].append(SimpleID(get_bone_name(bone), bone))
            if 'children' not in node:
                node['children'] = []
            offset = len(node['children'])
            root_bones = [bone for bone in blender_data.data.bones if bone.parent is None]
            root_refs = [
                Reference('objects', get_bone_name(b), node['children'], i + offset)
                for i, b in enumerate(root_bones)
            ]

            for bone in root_refs:
                state['references'].append(bone)
            node['children'].extend(root_refs)

        (
            node['translation'],
            node['rotation'],
            node['scale']
        ) = decompose(blender_data.matrix_local)

        return node

    @classmethod
    def export_dupli_group(cls, state, dupli_group):
        group_sid = SimpleID(
            'dupli_group_{}.{}'.format(dupli_group.name, len(state['dupli_nodes']))
        )
        state['input']['dupli_ids'].append(group_sid)

        group_node = {
            'name': group_sid.name,
        }
        state['dupli_nodes'].append(group_node)

        dupli_nodes = [group_node]
        children = []

        for dupli_obj in dupli_group.objects:
            obj_node = cls.export(state, dupli_obj)
            state['dupli_nodes'].append(obj_node)

            obj_sid = SimpleID(
                'dupli_node_{}.{}'.format(dupli_obj.name, len(state['dupli_nodes']))
            )
            state['input']['dupli_ids'].append(obj_sid)

            obj_node['name'] = obj_sid.name
            if dupli_obj.dupli_group is None and 'children' in obj_node:
                del obj_node['children']

            obj_ref = Reference('objects', obj_sid.name, children, len(children))
            children.append(obj_ref)
            state['references'].append(obj_ref)

            dupli_nodes.append(obj_node)

        group_node['children'] = children

        group_ref = Reference('objects', group_sid.name, None, None)
        state['references'].append(group_ref)
        return group_ref



import glob
import imp
import importlib
import os.path

FILES = [
    os.path.basename(f)[:-3]
    for f in glob.glob(os.path.dirname(__file__) + '/*.py')
    if os.path.isfile(f)
]
MODULES = [f for f in FILES if not f.startswith('_')]

__all__ = []
for module in MODULES:
    module = importlib.import_module('.'+module, __name__)
    if '_IMPORTED' in locals():
        imp.reload(module)
    for attr in [getattr(module, attr) for attr in dir(module)]:
        if hasattr(attr, 'ext_meta'):
            __all__.append(attr)

if '_IMPORTED' not in locals():
    _IMPORTED = True



from collections import namedtuple

Settings = namedtuple('Settings', (
    'material_mode',
    'blender_dir',
))
Settings.__new__.__defaults__ = (
    'legacy', # material_mode
    '', # blender_dir
)

class ConverterBase:
    '''Implements common functionality for converters'''

    def __init__(self, settings=None):
        if settings is None:
            settings = Settings(
            )
        self.settings = settings

    def convert_single(self, src, dst):
        '''Convert a single src file to dst'''
        raise NotImplementedError()

    def convert_batch(self, srcroot, dstdir, files):
        '''Convert files from srcroot to dstdir'''
        raise NotImplementedError()



from distutils.version import StrictVersion as Version

import mathutils

from .base import BaseExporter
from .common import (
    Buffer,
    Reference,
    SimpleID,
    get_bone_name,
)


def togl(matrix):
    return [i for col in matrix.col for i in col]


class SkinExporter(BaseExporter):
    gltf_key = 'skins'
    blender_key = 'skins'

    @classmethod
    def export(cls, state, blender_data):
        if state['version'] < Version('2.0'):
            joints_key = 'jointNames'
        else:
            joints_key = 'joints'

        arm = blender_data.find_armature()

        axis_mat = mathutils.Matrix.Identity(4)
        if state['settings']['nodes_global_matrix_apply']:
            axis_mat = state['settings']['nodes_global_matrix']

        bind_shape_mat = (
            axis_mat
            * arm.matrix_world.inverted()
            * blender_data.matrix_world
            * axis_mat.inverted()
        )

        bone_groups = [
            group for group in blender_data.vertex_groups if group.name in arm.data.bones
        ]

        gltf_skin = {
            'name': blender_data.name,
        }
        gltf_skin[joints_key] = [
            Reference('objects', get_bone_name(arm.data.bones[group.name]), None, None)
            for group in bone_groups
        ]
        for i, ref in enumerate(gltf_skin[joints_key]):
            ref.source = gltf_skin[joints_key]
            ref.prop = i
            state['references'].append(ref)

        if state['version'] < Version('2.0'):
            gltf_skin['bindShapeMatrix'] = togl(mathutils.Matrix.Identity(4))
        else:
            bone_names = [get_bone_name(b) for b in arm.data.bones if b.parent is None]
            if len(bone_names) > 1:
                print('Warning: Armature {} has no root node'.format(arm.data.name))
            gltf_skin['skeleton'] = Reference('objects', bone_names[0], gltf_skin, 'skeleton')
            state['references'].append(gltf_skin['skeleton'])

        element_size = 16 * 4
        num_elements = len(bone_groups)
        buf = Buffer('IBM_{}_skin'.format(blender_data.name))
        buf_view = buf.add_view(element_size * num_elements, element_size, None)
        idata = buf.add_accessor(buf_view, 0, element_size, Buffer.FLOAT, num_elements, Buffer.MAT4)

        for i, group in enumerate(bone_groups):
            bone = arm.data.bones[group.name]
            mat = togl((axis_mat * bone.matrix_local).inverted() * bind_shape_mat)
            for j in range(16):
                idata[(i * 16) + j] = mat[j]

        gltf_skin['inverseBindMatrices'] = Reference(
            'accessors',
            idata.name,
            gltf_skin,
            'inverseBindMatrices'
        )
        state['references'].append(gltf_skin['inverseBindMatrices'])
        state['buffers'].append(buf)
        state['input']['buffers'].append(SimpleID(buf.name))

        return gltf_skin



from distutils.version import StrictVersion as Version

import mathutils

from .base import BaseExporter
from .common import (
    Buffer,
    Reference,
    SimpleID,
)


OES_ELEMENT_INDEX_UINT = 'OES_element_index_uint'


class OffsetTracker:
    def __init__(self):
        self.value = 0

    def get(self):
        return self.value

    def add(self, value):
        self.value += value


class Vertex:
    __slots__ = (
        "co",
        "normal",
        "uvs",
        "colors",
        "loop_indices",
        "index",
        "weights",
        "joint_indexes",
        )

    def __init__(self, mesh, loop):
        vert_idx = loop.vertex_index
        loop_idx = loop.index
        self.co = mesh.vertices[vert_idx].co[:]
        self.normal = loop.normal[:]
        self.uvs = tuple(layer.data[loop_idx].uv[:] for layer in mesh.uv_layers)
        self.colors = tuple(layer.data[loop_idx].color[:] for layer in mesh.vertex_colors)
        self.loop_indices = [loop_idx]

        # Take the four most influential groups
        groups = sorted(
            mesh.vertices[vert_idx].groups,
            key=lambda group: group.weight,
            reverse=True
        )
        if len(groups) > 4:
            groups = groups[:4]

        self.weights = [group.weight for group in groups]
        self.joint_indexes = [group.group for group in groups]

        if len(self.weights) < 4:
            for _ in range(len(self.weights), 4):
                self.weights.append(0.0)
                self.joint_indexes.append(0)

        self.index = 0

    def __hash__(self):
        return hash((self.co, self.normal, self.uvs, self.colors))

    def __eq__(self, other):
        equals = (
            (self.co == other.co) and
            (self.normal == other.normal) and
            (self.uvs == other.uvs) and
            (self.colors == other.colors)
        )

        if equals:
            indices = self.loop_indices + other.loop_indices
            self.loop_indices = indices
            other.loop_indices = indices
        return equals


def requires_int_indices(vert_list):
    return len(vert_list) > 65535


def get_vert_list(mesh, has_shape_keys):
    mesh.calc_normals_split()
    mesh.calc_tessface()

    # Remove duplicate verts with dictionary hashing (causes problems with shape keys)
    if has_shape_keys:
        vert_list = [Vertex(mesh, loop) for loop in mesh.loops]
    else:
        vert_list = list({Vertex(mesh, loop): 0 for loop in mesh.loops}.keys())

    return vert_list


def gather_primitives(state, mesh, vert_lists):
    # For each material, make an empty primitive set.
    # This dictionary maps material names to list of indices that form the
    # part of the mesh that the material should be applied to.
    mesh_materials = [ma for ma in mesh.materials if ma in state['input']['materials']]
    prims = {ma.name if ma else '': [] for ma in mesh_materials}
    if not prims:
        prims = {'': []}

    # Index data
    # Map loop indices to vertices
    vert_dict = {i: vertex for vertex in vert_lists[0] for i in vertex.loop_indices}

    for poly in mesh.polygons:
        # Find the primitive that this polygon ought to belong to (by
        # material).
        if not mesh_materials:
            prim = prims['']
        else:
            try:
                mat = mesh_materials[poly.material_index]
            except IndexError:
                # Polygon has a bad material index, so skip it
                continue
            prim = prims[mat.name if mat else '']

        # Find the (vertex) index associated with each loop in the polygon.
        indices = [vert_dict[i].index for i in poly.loop_indices]
        coords = [mathutils.Vector(vert_dict[i].co) for i in poly.loop_indices]

        if len(indices) == 3:
            # No triangulation necessary
            prim += indices
        elif len(indices) > 3:
            # Triangulation necessary
            triangles = mathutils.geometry.tessellate_polygon((coords,))
            for triangle in triangles:
                prim += [indices[i] for i in triangle[::-1]]
        else:
            # Bad polygon
            raise RuntimeError(
                "Invalid polygon with {} vertices.".format(len(indices))
            )
    return prims


def fill_data(acc, iterator):
    for i, data in enumerate(iterator):
        size = len(data)
        for j, element in enumerate(data):
            acc[(i * size + j)] = element


class MeshExporter(BaseExporter):
    gltf_key = 'meshes'
    blender_key = 'meshes'

    @classmethod
    def export(cls, state, blender_data):
        # glTF data
        mesh_name = blender_data.name
        blender_data = state['mod_meshes'].get(blender_data.name, blender_data)
        gltf_mesh = {
            'name': mesh_name,
            'primitives': [],
        }

        extras = BaseExporter.get_custom_properties(blender_data)
        if extras:
            gltf_mesh['extras'] = extras

        shape_keys = state['shape_keys'].get(mesh_name, [])

        vert_lists = [get_vert_list(blender_data, len(shape_keys) > 0)]
        vert_lists += [
            get_vert_list(key_mesh, True)
            for key_mesh in [key[1] for key in shape_keys]
        ]

        # Process mesh data and gather attributes
        buffer = Buffer(mesh_name)
        gltf_attrs = cls.export_attributes(state, buffer, mesh_name, None, vert_lists[0])

        # Process shape keys
        targets = [
            cls.export_attributes(
                state,
                Buffer(key_mesh.name),
                key_mesh.name,
                vert_lists[0],
                vert_lists[i + 1]
            )
            for i, key_mesh in enumerate([key[1] for key in shape_keys])
        ]

        if shape_keys:
            gltf_mesh['weights'] = [key[0] for key in shape_keys]

        prims = gather_primitives(state, blender_data, vert_lists)

        if requires_int_indices(vert_lists[0]):
            # Use the integer index extension
            if OES_ELEMENT_INDEX_UINT not in state['gl_extensions_used']:
                state['gl_extensions_used'].append(OES_ELEMENT_INDEX_UINT)
            itype = Buffer.UNSIGNED_INT
        else:
            itype = Buffer.UNSIGNED_SHORT

        gltf_mesh['primitives'] = [
            cls.export_primitive(state, buffer, mat, indices, itype, gltf_attrs, targets)
            for mat, indices in prims.items() if indices
        ]

        return gltf_mesh

    @classmethod
    def export_primitive(cls, state, buf, material, indices, index_type, attributes, targets):
        index_stride = 4 if index_type == Buffer.UNSIGNED_INT else 2

        # Pad index buffer if necessary to maintain a size that is a multiple of 4
        view_length = index_stride * len(indices)
        view_length = view_length + ((4 - view_length % 4) % 4)

        index_view = buf.add_view(view_length, 0, Buffer.ELEMENT_ARRAY_BUFFER)
        idata = buf.add_accessor(
            index_view,
            0,
            index_stride,
            index_type,
            len(indices),
            Buffer.SCALAR
        )

        for i, index in enumerate(indices):
            idata[i] = index

        gltf_prim = {
            'attributes': attributes,
            'mode': 4,
        }

        gltf_prim['indices'] = Reference('accessors', idata.name, gltf_prim, 'indices')
        state['references'].append(gltf_prim['indices'])

        if targets:
            gltf_prim['targets'] = targets

        # Add the material reference after checking that it is valid
        if material:
            gltf_prim['material'] = Reference('materials', material, gltf_prim, 'material')
            state['references'].append(gltf_prim['material'])

        return gltf_prim

    @classmethod
    def export_attributes(cls, state, buf, mesh_name, base_vert_list, vert_list):
        gltf_attrs = {}
        is_skinned = mesh_name in state['skinned_meshes']

        color_size = 4 if state['settings']['meshes_vertex_color_alpha'] else 3
        num_uv_layers = len(vert_list[0].uvs)
        num_col_layers = len(vert_list[0].colors)
        vertex_size = (3 + 3 + num_uv_layers * 2 + num_col_layers * color_size) * 4
        if is_skinned:
            vertex_size += (4 + 4) * 4
        num_verts = len(vert_list)

        if state['settings']['meshes_interleave_vertex_data']:
            view = buf.add_view(vertex_size * num_verts, vertex_size, Buffer.ARRAY_BUFFER)
        else:
            view = None

        offset = OffsetTracker()

        def create_attr_accessor(name, component_type, component_count):
            if not view:
                stride = 4 * component_count
                buffer = Buffer(mesh_name + '_' + name)
                state['buffers'].append(buffer)
                state['input']['buffers'].append(SimpleID(buffer.name))
                _view = buffer.add_view(stride * num_verts, stride, Buffer.ARRAY_BUFFER)
                interleaved = False
            else:
                _view = view
                buffer = buf
                stride = vertex_size
                interleaved = True

            data_type = [Buffer.SCALAR, Buffer.VEC2, Buffer.VEC3, Buffer.VEC4][component_count - 1]
            _offset = offset.get()
            acc = buffer.add_accessor(_view, _offset, stride, component_type, num_verts, data_type)
            if interleaved:
                offset.add(4 * component_count)
            return acc

        def add_attribute(name, component_size, component_type=Buffer.FLOAT):
            acc = create_attr_accessor(name, component_type, component_size)
            gltf_attrs[name] = Reference('accessors', acc.name, gltf_attrs, name)
            state['references'].append(gltf_attrs[name])
            return acc

        vdata = add_attribute('POSITION', 3)
        ndata = add_attribute('NORMAL', 3)
        if not base_vert_list:
            tdata = [
                add_attribute('TEXCOORD_' + str(i), 2)
                for i in range(num_uv_layers)
            ]
            cdata = [
                add_attribute('COLOR_' + str(i), color_size)
                for i in range(num_col_layers)
            ]

        # Copy vertex data
        if base_vert_list:
            fill_data(vdata, (
                [a - b for a, b in zip(v.co, b.co)]
                for v, b in zip(vert_list, base_vert_list)
            ))
            fill_data(ndata, (
                [a - b for a, b in zip(v.normal, b.normal)]
                for v, b in zip(vert_list, base_vert_list)
            ))
        else:
            fill_data(vdata, (v.co for v in vert_list))
            fill_data(ndata, (v.normal for v in vert_list))
            for i, accessor in enumerate(cdata):
                if state['settings']['meshes_vertex_color_alpha']:
                    fill_data(accessor, (v.colors[i] + [1.0] for v in vert_list))
                else:
                    fill_data(accessor, (v.colors[i] for v in vert_list))
            for i, accessor in enumerate(tdata):
                if state['settings']['asset_profile'] == 'WEB':
                    fill_data(accessor, ((v.uvs[i][0], 1.0 - v.uvs[i][1]) for v in vert_list))
                else:
                    fill_data(accessor, (v.uvs[i] for v in vert_list))
            for i, vtx in enumerate(vert_list):
                vtx.index = i

        state['buffers'].append(buf)
        state['input']['buffers'].append(SimpleID(buf.name))

        if is_skinned:
            if state['version'] < Version('2.0'):
                joint_key = 'JOINT'
                weight_key = 'WEIGHT'
            else:
                joint_key = 'JOINTS_0'
                weight_key = 'WEIGHTS_0'

            jdata = add_attribute(joint_key, 4, Buffer.UNSIGNED_BYTE)
            fill_data(jdata, (v.joint_indexes for v in vert_list))
            wdata = add_attribute(weight_key, 4)
            fill_data(wdata, (v.weights for v in vert_list))

        return gltf_attrs



import base64
import os
import struct
import zlib

import bpy

from .base import BaseExporter
from .common import (
    Buffer,
    Reference,
    SimpleID,
)


EXT_MAP = {'BMP': 'bmp', 'JPEG': 'jpg', 'PNG': 'png', 'TARGA': 'tga'}


class ImageExporter(BaseExporter):
    gltf_key = 'images'
    blender_key = 'images'

    @classmethod
    def check(cls, state, blender_data):
        errors = []
        if blender_data.size[0] == 0:
            errors.append('x dimension is 0')
        if blender_data.size[1] == 0:
            errors.append('y dimension is 0')
        if blender_data.type != 'IMAGE':
            errors.append('not an image {}'.format(blender_data.type))

        if errors:
            err_list = '\n\t'.join(errors)
            print(
                'Unable to export image {} due to the following errors:\n\t{}'
                .format(blender_data.name, err_list)
            )
            return False

        return True

    @classmethod
    def default(cls, state, blender_data):
        return {
            'name': blender_data.name,
            'uri': ''
        }

    @classmethod
    def image_to_data_uri(cls, image):
        width = image.size[0]
        height = image.size[1]
        buf = bytearray([int(p * 255) for p in image.pixels])

        # reverse the vertical line order and add null bytes at the start
        width_byte_4 = width * 4
        raw_data = b''.join(b'\x00' + buf[span:span + width_byte_4]
                            for span in range((height - 1) * width_byte_4, -1, - width_byte_4))

        def png_pack(png_tag, data):
            chunk_head = png_tag + data
            return (struct.pack("!I", len(data)) +
                    chunk_head +
                    struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head)))

        png_bytes = b''.join([
            b'\x89PNG\r\n\x1a\n',
            png_pack(b'IHDR', struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
            png_pack(b'IDAT', zlib.compress(raw_data, 9)),
            png_pack(b'IEND', b'')])

        return png_bytes

    @classmethod
    def export(cls, state, blender_data):
        path = ''
        data = None

        gltf = {'name': blender_data.name}

        storage_setting = state['settings']['images_data_storage']
        image_packed = blender_data.packed_file is not None
        if image_packed and storage_setting in ['COPY', 'REFERENCE']:
            if blender_data.file_format in EXT_MAP:
                # save the file to the output directory
                gltf['uri'] = '.'.join([blender_data.name, EXT_MAP[blender_data.file_format]])
                temp = blender_data.filepath
                blender_data.filepath = os.path.join(
                    state['settings']['gltf_output_dir'],
                    gltf['uri']
                )
                blender_data.save()
                with open(bpy.path.abspath(blender_data.filepath), 'rb') as fin:
                    data = fin.read()
                blender_data.filepath = temp
            else:
                # convert to png and save
                gltf['uri'] = '.'.join([blender_data.name, 'png'])
                data = cls.image_to_data_uri(blender_data)
            path = os.path.join(state['settings']['gltf_output_dir'], gltf['uri'])

        elif storage_setting == 'COPY':
            with open(bpy.path.abspath(blender_data.filepath), 'rb') as fin:
                data = fin.read()
            gltf['uri'] = bpy.path.basename(blender_data.filepath)
            path = os.path.join(state['settings']['gltf_output_dir'], gltf['uri'])
        elif storage_setting == 'REFERENCE':
            gltf['uri'] = blender_data.filepath.replace('//', '')
        elif storage_setting == 'EMBED':
            png_bytes = cls.image_to_data_uri(blender_data)
            gltf['mimeType'] = 'image/png'
            if state['settings']['gltf_export_binary']:
                buf = Buffer(blender_data.name)
                view_key = buf.add_view(len(png_bytes), 0, None)
                view = buf.buffer_views[view_key]
                view['data'] = png_bytes

                pad = 4 - len(png_bytes) % 4
                if pad not in [0, 4]:
                    buf.add_view(pad, 0, None)

                gltf['bufferView'] = Reference('bufferViews', view_key, gltf, 'bufferView')
                state['references'].append(gltf['bufferView'])

                state['buffers'].append(buf)
                state['input']['buffers'].append(SimpleID('buffer_' + blender_data.name))
            else:
                gltf['uri'] = 'data:image/png;base64,' + base64.b64encode(png_bytes).decode()
        else:
            print(
                'Encountered unknown option ({}) for images_data_storage setting'
                .format(storage_setting)
            )

        if path:
            state['files'][path] = data

        return gltf



from .base import BaseExporter
from .common import (
    Reference,
)


class SceneExporter(BaseExporter):
    gltf_key = 'scenes'
    blender_key = 'scenes'

    @classmethod
    def export(cls, state, blender_data):
        bg_color = blender_data.world.horizon_color[:] if blender_data.world else [0.05]*3
        result = {
            'extras': {
                'background_color': bg_color,
                'frames_per_second': blender_data.render.fps,
            },
            'name': blender_data.name,
        }

        if blender_data.camera and blender_data.camera.data in state['input']['cameras']:
            result['extras']['active_camera'] = Reference(
                'cameras',
                blender_data.camera.name,
                result['extras'],
                'active_camera'
            )
            state['references'].append(result['extras']['active_camera'])

        extras = BaseExporter.get_custom_properties(blender_data)
        if extras:
            result['extras'].update(BaseExporter.get_custom_properties(blender_data))

        result['nodes'] = [
            Reference('objects', ob.name, None, None)
            for ob in blender_data.objects
            if ob in state['input']['objects'] and ob.parent is None and ob.is_visible(blender_data)
        ]
        for i, ref in enumerate(result['nodes']):
            ref.source = result['nodes']
            ref.prop = i
        state['references'].extend(result['nodes'])

        hidden_nodes = [
            Reference('objects', ob.name, None, None)
            for ob in blender_data.objects
            if ob in state['input']['objects'] and not ob.is_visible(blender_data)
        ]

        if hidden_nodes:
            result['extras']['hidden_nodes'] = hidden_nodes
            for i, ref in enumerate(hidden_nodes):
                ref.source = result['extras']['hidden_nodes']
                ref.prop = i
            state['references'].extend(result['extras']['hidden_nodes'])

        return result



from .base import BaseExporter

from .common import (
    Buffer,
    Reference,
    SimpleID,
    get_bone_name,
)

from .animation import AnimationExporter
from .camera import CameraExporter
from .image import ImageExporter
from .material import MaterialExporter
from .mesh import MeshExporter
from .node import NodeExporter
from .scene import SceneExporter
from .skin import SkinExporter
from .texture import TextureExporter



import json
import os
import sys

import panda3d.core as p3d

from blend2bam.common import ConverterBase

class ConverterGltf2Bam(ConverterBase):
    def convert_single(self, src, dst):
        scriptdir = os.path.dirname(os.path.abspath(__file__))
        sys.path.insert(0, os.path.join(scriptdir, 'panda3d-gltf'))
        from gltf.converter import Converter #pylint: disable=import-error

        dstdir = os.path.dirname(dst)
        os.makedirs(dstdir, exist_ok=True)

        with open(src) as gltf_file:
            gltf_data = json.load(gltf_file)

        dstfname = p3d.Filename.fromOsSpecific(dst)
        p3d.get_model_path().prepend_directory(dstfname.getDirname())

        indir = p3d.Filename(p3d.Filename.from_os_specific(src).get_dirname())
        outdir = p3d.Filename(dstfname.get_dirname())

        converter = Converter(indir=indir, outdir=outdir)
        converter.update(gltf_data, writing_bam=True)
        converter.active_scene.write_bam_file(dstfname)

    def convert_batch(self, srcroot, dstdir, files):
        for gltffile in files:
            src = gltffile
            dst = src.replace(srcroot, dstdir).replace('.gltf', '.bam')

            self.convert_single(src, dst)



from .converter import load_model


def patch_loader(loader):
    '''Monkey patch the supplied Loader to add glTF support'''
    import types
    import panda3d.core as p3d
    _load_model = loader.load_model

    def new_load_model(self, model_path, **kwargs):
        fname = p3d.Filename(model_path)
        if fname.get_extension() in 'gltf, glb':
            return load_model(self, model_path, **kwargs)
        else:
            return _load_model(model_path, **kwargs)
    loader.load_model = loader.loadModel = types.MethodType(new_load_model, loader)


__all__ = [
    'load_model',
]



from blendergltf.exporters.common import Reference #pylint: disable=import-error,wrong-import-position


class ExtMaterialsLegacy:
    ext_meta = {
        'name': 'BP_materials_legacy',
        'isDraft': True,
    }

    def export_material(self, state, material):
        all_textures = [
            slot for slot in material.texture_slots
            if slot and slot.texture.type == 'IMAGE'
        ]
        diffuse_textures = [
            Reference('textures', t.texture.name, t.texture, None)
            for t in all_textures if t.use_map_color_diffuse
        ]
        emission_textures = [
            Reference('textures', t.texture.name, t.texture, None)
            for t in all_textures
            if (
                (material.use_shadeless and t.use_map_color_diffuse)
                or (not material.use_shadeless and t.use_map_emit)
            )
        ]
        specular_textures = [
            Reference('textures', t.texture.name, t.texture, None)
            for t in all_textures if t.use_map_color_spec
        ]

        diffuse_color = list((material.diffuse_color * material.diffuse_intensity)[:])
        diffuse_color += [material.alpha]
        emission_color = list((material.diffuse_color * material.emit)[:])
        emission_color += [material.alpha]
        specular_color = list((material.specular_color * material.specular_intensity)[:])
        specular_color += [material.specular_alpha]

        gltf = {
            'bpLegacy': {
                'diffuseFactor': diffuse_color,
                'emissionFactor': emission_color,
                'specularFactor': specular_color,
                'ambientFactor': ([material.ambient]*3) + [1.0],
                'shininessFactor': material.specular_hardness,
            }
        }

        def handle_tex(name, tex):
            slotname = '{}Texture'.format(name)
            gltf['bpLegacy'][slotname] = {
                'index': tex,
                'texCoord': 0, # FIXME
            }

            use_srgb = tex.source.image.colorspace_settings.name == 'sRGB'
            gltf['bpLegacy']['{}Srgb'.format(slotname)] = use_srgb

            tex.source = gltf['bpLegacy'][slotname]
            tex.prop = 'index'
            state['references'].append(tex)


        if diffuse_textures:
            handle_tex('diffuse', diffuse_textures[-1])
        if emission_textures:
            handle_tex('emission', emission_textures[-1])
        if specular_textures:
            handle_tex('specular', specular_textures[-1])

        return gltf

    def export(self, state):
        state['extensions_used'].append('BP_materials_legacy')

        # Export materials
        material_pairs = [
            (material, state['output']['materials'][state['refmap'][('materials', material.name)]])
            for material in state['input']['materials']
        ]
        for bl_mat, gl_mat in material_pairs:
            gl_mat['extensions'] = gl_mat.get('extensions', {})
            gl_mat['extensions']['BP_materials_legacy'] = self.export_material(state, bl_mat)


class ExtZup:
    ext_meta = {
        'name': 'BP_zup',
    }

    def export(self, state):
        state['extensions_used'].append('BP_zup')



import os
import json
import shutil
import tempfile

from blend2bam import blenderutils
from blend2bam.common import ConverterBase

class ConverterBlend2Gltf(ConverterBase):
    script_file = os.path.join(os.path.dirname(__file__), 'blender_script.py')

    def convert_single(self, src, dst):
        srcroot = os.path.dirname(src)
        dstdir = os.path.dirname(dst)
        files = [
            src
        ]
        self.convert_batch(srcroot, dstdir, files)

        dstout = os.path.join(dstdir, os.path.basename(src).replace('.blend', '.gltf'))
        shutil.move(dstout, dst)

    def convert_batch(self, srcroot, dstdir, files):
        blenderdir = self.settings.blender_dir
        settings_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        json.dump(self.settings._asdict(), settings_file)
        settings_file.close() # Close so the tempfile can be re-opened in Blender on Windows
        args = [
            settings_file.name,
            srcroot,
            dstdir,
        ] + files
        blenderutils.run_blender_script(self.script_file, args, blenderdir=blenderdir)
        os.remove(settings_file.name)



from distutils.version import StrictVersion as Version

from .base import BaseExporter
from .common import (
    Reference,
)


class MaterialExporter(BaseExporter):
    gltf_key = 'materials'
    blender_key = 'materials'

    @classmethod
    def export(cls, state, blender_data):
        gltf = {
            'name': blender_data.name,
        }

        if state['version'] < Version('2.0'):
            return gltf

        if hasattr(blender_data, 'pbr_export_settings'):
            gltf['doubleSided'] = not blender_data.game_settings.use_backface_culling
            pbr_settings = blender_data.pbr_export_settings
            pbr = {
                'baseColorFactor': pbr_settings.base_color_factor[:],
                'metallicFactor': pbr_settings.metallic_factor,
                'roughnessFactor': pbr_settings.roughness_factor,
            }

            gltf['alphaMode'] = pbr_settings.alpha_mode
            if gltf['alphaMode'] == 'MASK':
                gltf['alphaCutoff'] = pbr_settings.alpha_cutoff

            input_textures = [texture.name for texture in state['input']['textures']]
            base_color_text = pbr_settings.base_color_texture
            if base_color_text and base_color_text in input_textures:
                pbr['baseColorTexture'] = {
                    'texCoord': pbr_settings.base_color_text_index,
                }
                pbr['baseColorTexture']['index'] = Reference(
                    'textures',
                    pbr_settings.base_color_texture,
                    pbr['baseColorTexture'],
                    'index'
                )
                state['references'].append(pbr['baseColorTexture']['index'])

            metal_rough_text = pbr_settings.metal_roughness_texture
            if metal_rough_text and metal_rough_text in input_textures:
                pbr['metallicRoughnessTexture'] = {
                    'texCoord': pbr_settings.metal_rough_text_index,
                }
                pbr['metallicRoughnessTexture']['index'] = Reference(
                    'textures',
                    pbr_settings.metal_roughness_texture,
                    pbr['metallicRoughnessTexture'],
                    'index'
                )
                state['references'].append(pbr['metallicRoughnessTexture']['index'])

            gltf['pbrMetallicRoughness'] = pbr

            gltf['emissiveFactor'] = pbr_settings.emissive_factor[:]

            emissive_text = pbr_settings.emissive_texture
            if emissive_text and emissive_text in input_textures:
                gltf['emissiveTexture'] = {
                    'texCoord': pbr_settings.emissive_text_index,
                }
                gltf['emissiveTexture']['index'] = Reference(
                    'textures',
                    pbr_settings.emissive_texture,
                    gltf['emissiveTexture'],
                    'index'
                )
                state['references'].append(gltf['emissiveTexture']['index'])

            normal_text = pbr_settings.normal_texture
            if normal_text and normal_text in input_textures:
                gltf['normalTexture'] = {
                    'texCoord': pbr_settings.normal_text_index,
                }
                gltf['normalTexture']['index'] = Reference(
                    'textures',
                    pbr_settings.normal_texture,
                    gltf['normalTexture'],
                    'index'
                )
                state['references'].append(gltf['normalTexture']['index'])

            occlusion_text = pbr_settings.occlusion_texture
            if occlusion_text and occlusion_text in input_textures:
                gltf['occlusionTexture'] = {
                    'texCoord': pbr_settings.occlusion_text_index,
                }
                gltf['occlusionTexture']['index'] = Reference(
                    'textures',
                    pbr_settings.occlusion_texture,
                    gltf['occlusionTexture'],
                    'index'
                )
                state['references'].append(gltf['occlusionTexture']['index'])

        return gltf
